# This script extracts the execution time for
# various different settings of tsfresh
# using different input data
# Attention: it will run for ~half a day
# Do these calculations in a controlled environment
# (e.g. a cloud provider VM)
# You will need to have b2luigi installed.
from tsfresh.feature_extraction import ComprehensiveFCParameters, MinimalFCParameters, extract_features

import pandas as pd
import numpy as np
from time import time
import b2luigi as luigi
import json


class DataCreationTask(luigi.Task):
    """Create random data for testing"""
    num_ids = luigi.IntParameter(default=100)
    time_series_length = luigi.IntParameter()
    random_seed = luigi.IntParameter()

    def output(self):
        yield self.add_to_output("data.csv")

    def run(self):
        np.random.seed(self.random_seed)

        df = pd.concat([
            pd.DataFrame({
                "id": [i] * self.time_series_length,
                "time": range(self.time_series_length),
                "value": np.random.randn(self.time_series_length)
            })
            for i in range(self.num_ids)
        ])

        with self._get_output_target("data.csv").open("w") as f:
            df.to_csv(f)


@luigi.requires(DataCreationTask)
class TimingTask(luigi.Task):
    """Run tsfresh with the given parameters"""
    feature_parameter = luigi.DictParameter(hashed=True)
    n_jobs = luigi.IntParameter()
    try_number = luigi.IntParameter()

    def output(self):
        yield self.add_to_output("result.json")

    def run(self):
        input_file = self._get_input_targets("data.csv")[0]

        with input_file.open("r") as f:
            df = pd.read_csv(f)

        start_time = time()
        extract_features(df, column_id="id", column_sort="time", n_jobs=self.n_jobs,
                         default_fc_parameters=self.feature_parameter,
                         disable_progressbar=True)
        end_time = time()

        single_parameter_name = list(self.feature_parameter.keys())[0]
        single_parameter_params = self.feature_parameter[single_parameter_name]

        result_json = {
            "time": end_time - start_time,
            "n_ids": self.num_ids,
            "n_jobs": self.n_jobs,
            "feature": single_parameter_name,
            "number_parameters": len(single_parameter_params) if single_parameter_params else 0,
            "time_series_length": int((df["id"] == 0).sum()),
            "try_number": self.try_number,
        }

        with self._get_output_target("result.json").open("w") as f:
            json.dump(result_json, f)


@luigi.requires(DataCreationTask)
class FullTimingTask(luigi.Task):
    """Run tsfresh with all calculators for comparison"""
    n_jobs = luigi.IntParameter()

    def output(self):
        yield self.add_to_output("result.json")

    def run(self):
        input_file = self._get_input_targets("data.csv")[0]

        with input_file.open("r") as f:
            df = pd.read_csv(f)

        start_time = time()
        extract_features(df, column_id="id", column_sort="time", n_jobs=self.n_jobs,
                         disable_progressbar=True)
        end_time = time()

        result_json = {
            "time": end_time - start_time,
            "n_ids": self.num_ids,
            "n_jobs": self.n_jobs,
            "time_series_length": int((df["id"] == 0).sum()),
        }

        with self._get_output_target("result.json").open("w") as f:
            json.dump(result_json, f)


class CombinerTask(luigi.Task):
    """Collect all tasks into a single result.csv file"""
    def complete(self):
        return False

    def requires(self):
        settings = ComprehensiveFCParameters()
        for job in [0, 1, 4]:
            for time_series_length in [100, 500, 1000, 5000]:
                yield FullTimingTask(time_series_length=time_series_length,
                                     n_jobs=job,
                                     num_ids=10,
                                     random_seed=42)
                yield FullTimingTask(time_series_length=time_series_length,
                                     n_jobs=job,
                                     num_ids=100,
                                     random_seed=42)

                for feature_name in settings:
                    yield TimingTask(
                        feature_parameter={feature_name: settings[feature_name]},
                        time_series_length=time_series_length,
                        n_jobs=job,
                        num_ids=100,
                        try_number=0,
                        random_seed=42
                    )

                    for try_number in range(3):
                        yield TimingTask(
                            feature_parameter={feature_name: settings[feature_name]},
                            n_jobs=job,
                            try_number=try_number,
                            num_ids=10,
                            time_series_length=time_series_length,
                            random_seed=42
                        )

    def output(self):
        yield self.add_to_output("results.csv")

    def run(self):
        results = []

        for input_file in self._get_input_targets("result.json"):
            with input_file.open("r") as f:
                results.append(json.load(f))

        df = pd.DataFrame(results)

        with self._get_output_target("results.csv").open("w") as f:
            df.to_csv(f)


if __name__ == "__main__":
    luigi.set_setting("result_path", "results")
    luigi.process(CombinerTask())



# Testcase for benchmarking tsfresh feature extraction and selection
import pytest
import pandas as pd
import numpy as np

from tsfresh import extract_features, extract_relevant_features
from tsfresh.feature_extraction.settings import ComprehensiveFCParameters, MinimalFCParameters


def create_data(time_series_length, num_ids, random_seed=42):
    np.random.seed(random_seed)

    df = pd.concat([
        pd.DataFrame({
            "id": [i] * time_series_length,
            "time": range(time_series_length),
            "value": np.random.randn(time_series_length).cumsum()
        })
        for i in range(num_ids)
    ])

    return df


def test_benchmark_small_data(benchmark):
    df = create_data(5, 200)

    benchmark(extract_features, df, column_id="id", column_sort="time", n_jobs=0,
              disable_progressbar=True)


def test_benchmark_large_data(benchmark):
    df = create_data(500, 20)

    benchmark(extract_features, df, column_id="id", column_sort="time", n_jobs=0,
              disable_progressbar=True)


def test_benchmark_with_selection(benchmark):
    df = create_data(500, 20)
    y = pd.Series(np.random.choice([0, 1], 20))

    benchmark(extract_relevant_features, df, y, column_id="id", column_sort="time", n_jobs=0,
              disable_progressbar=True)



# -*- coding: utf-8 -*-
# This file as well as the whole tsfresh package are licenced under the MIT licence (see the LICENCE.txt)
# Maximilian Christ (maximilianchrist.com), Blue Yonder Gmbh, 2016

from random import shuffle
from unittest import TestCase
import warnings

from tsfresh.feature_extraction.feature_calculators import *
from tsfresh.feature_extraction.feature_calculators import _roll
from tsfresh.feature_extraction.feature_calculators import _get_length_sequences_where
from tsfresh.feature_extraction.feature_calculators import _estimate_friedrich_coefficients
from tsfresh.feature_extraction.feature_calculators import _aggregate_on_chunks
from tsfresh.feature_extraction.feature_calculators import _into_subchunks
from tsfresh.examples.driftbif_simulation import velocity
import math


class FeatureCalculationTestCase(TestCase):
    def setUp(self):
        # There will be a lot of warnings in the feature calculators.
        # Just ignore all of them in these tests
        warnings.simplefilter("ignore")

    def tearDown(self):
        warnings.resetwarnings()

    def assertIsNaN(self, result):
        self.assertTrue(np.isnan(result), msg="{} is not np.NaN")

    def assertEqualOnAllArrayTypes(self, f, input_to_f, result, *args, **kwargs):
        expected_result = f(input_to_f, *args, **kwargs)
        self.assertEqual(expected_result, result,
                         msg="Not equal for lists: {} != {}".format(expected_result, result))
        expected_result = f(np.array(input_to_f), *args, **kwargs)
        self.assertEqual(expected_result, result,
                         msg="Not equal for numpy.arrays: {} != {}".format(expected_result, result))
        expected_result = f(pd.Series(input_to_f, dtype="float64"), *args, **kwargs)
        self.assertEqual(expected_result, result,
                         msg="Not equal for pandas.Series: {} != {}".format(expected_result, result))

    def assertTrueOnAllArrayTypes(self, f, input_to_f, *args, **kwargs):
        self.assertTrue(f(input_to_f, *args, **kwargs), msg="Not true for lists")
        self.assertTrue(f(np.array(input_to_f), *args, **kwargs), msg="Not true for numpy.arrays")
        self.assertTrue(f(pd.Series(input_to_f), *args, **kwargs), msg="Not true for pandas.Series")

    def assertAllTrueOnAllArrayTypes(self, f, input_to_f, *args, **kwargs):
        self.assertTrue(all(dict(f(input_to_f, *args, **kwargs)).values()), msg="Not true for lists")
        self.assertTrue(all(dict(f(np.array(input_to_f), *args, **kwargs)).values()), msg="Not true for numpy.arrays")
        self.assertTrue(all(dict(f(pd.Series(input_to_f), *args, **kwargs)).values()), msg="Not true for pandas.Series")

    def assertFalseOnAllArrayTypes(self, f, input_to_f, *args, **kwargs):
        self.assertFalse(f(input_to_f, *args, **kwargs), msg="Not false for lists")
        self.assertFalse(f(np.array(input_to_f), *args, **kwargs), msg="Not false for numpy.arrays")
        self.assertFalse(f(pd.Series(input_to_f), *args, **kwargs), msg="Not false for pandas.Series")

    def assertAllFalseOnAllArrayTypes(self, f, input_to_f, *args, **kwargs):
        self.assertFalse(any(dict(f(input_to_f, *args, **kwargs)).values()), msg="Not false for lists")
        self.assertFalse(any(dict(f(np.array(input_to_f), *args, **kwargs)).values()),
                         msg="Not false for numpy.arrays")
        self.assertFalse(any(dict(f(pd.Series(input_to_f), *args, **kwargs)).values()),
                         msg="Not false for pandas.Series")

    def assertAlmostEqualOnAllArrayTypes(self, f, input_to_f, result, *args, **kwargs):
        expected_result = f(input_to_f, *args, **kwargs)
        self.assertAlmostEqual(expected_result, result,
                               msg="Not almost equal for lists: {} != {}".format(expected_result, result))
        expected_result = f(np.array(input_to_f), *args, **kwargs)
        self.assertAlmostEqual(expected_result, result,
                               msg="Not almost equal for numpy.arrays: {} != {}".format(expected_result, result))
        expected_result = f(pd.Series(input_to_f, dtype="float64"), *args, **kwargs)
        self.assertAlmostEqual(expected_result, result,
                               msg="Not almost equal for pandas.Series: {} != {}".format(expected_result, result))

    def assertIsNanOnAllArrayTypes(self, f, input_to_f, *args, **kwargs):
        self.assertTrue(np.isnan(f(input_to_f, *args, **kwargs)), msg="Not NaN for lists")
        self.assertTrue(np.isnan(f(np.array(input_to_f), *args, **kwargs)), msg="Not NaN for numpy.arrays")
        self.assertTrue(np.isnan(f(pd.Series(input_to_f, dtype="float64"), *args, **kwargs)),
                        msg="Not NaN for pandas.Series")

    def assertEqualPandasSeriesWrapper(self, f, input_to_f, result, *args, **kwargs):
        self.assertEqual(f(pd.Series(input_to_f), *args, **kwargs), result,
                         msg="Not equal for pandas.Series: {} != {}".format(
                             f(pd.Series(input_to_f), *args, **kwargs), result))

    def test__roll(self):
        x = np.random.normal(size=30)
        for shift in [0, 1, 10, 11, 30, 31, 50, 51, 150, 151]:
            np.testing.assert_array_equal(_roll(x, shift), np.roll(x, shift))
            np.testing.assert_array_equal(_roll(x, -shift), np.roll(x, -shift))

    def test___get_length_sequences_where(self):
        self.assertEqualOnAllArrayTypes(_get_length_sequences_where, [0, 1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 1],
                                        [1, 3, 1, 2])
        self.assertEqualOnAllArrayTypes(_get_length_sequences_where,
                                        [0, True, 0, 0, True, True, True, 0, 0, True, 0, True, True],
                                        [1, 3, 1, 2])
        self.assertEqualOnAllArrayTypes(_get_length_sequences_where,
                                        [0, True, 0, 0, 1, True, 1, 0, 0, True, 0, 1, True], [1, 3, 1, 2])
        self.assertEqualOnAllArrayTypes(_get_length_sequences_where, [0] * 10, [0])
        self.assertEqualOnAllArrayTypes(_get_length_sequences_where, [], [0])

    def test__into_subchunks(self):
        np.testing.assert_array_equal(_into_subchunks(range(7), 3, 2), np.array([[0, 1, 2], [2, 3, 4], [4, 5, 6]]))
        np.testing.assert_array_equal(_into_subchunks(range(5), 3), np.array([[0, 1, 2], [1, 2, 3], [2, 3, 4]]))

    def test_variance_larger_than_standard_deviation(self):
        self.assertFalseOnAllArrayTypes(variance_larger_than_standard_deviation, [-1, -1, 1, 1, 1])
        self.assertTrueOnAllArrayTypes(variance_larger_than_standard_deviation, [-1, -1, 1, 1, 2])

    def test_large_standard_deviation(self):
        self.assertFalseOnAllArrayTypes(large_standard_deviation, [1, 1, 1, 1], r=0)
        self.assertFalseOnAllArrayTypes(large_standard_deviation, [1, 1, 1, 1], r=0)
        self.assertTrueOnAllArrayTypes(large_standard_deviation, [-1, -1, 1, 1], r=0)
        self.assertTrueOnAllArrayTypes(large_standard_deviation, [-1, -1, 1, 1], r=0.25)
        self.assertTrueOnAllArrayTypes(large_standard_deviation, [-1, -1, 1, 1], r=0.3)
        self.assertFalseOnAllArrayTypes(large_standard_deviation, [-1, -1, 1, 1], r=0.5)

    def test_symmetry_looking(self):
        self.assertAllTrueOnAllArrayTypes(symmetry_looking, [-1, -1, 1, 1],
                                          [dict(r=0.05), dict(r=0.75)])
        self.assertAllFalseOnAllArrayTypes(symmetry_looking, [-1, -1, 1, 1], [dict(r=0)])
        self.assertAllFalseOnAllArrayTypes(symmetry_looking, [-1, -1, -1, -1, 1], [dict(r=0.05)])
        self.assertAllTrueOnAllArrayTypes(symmetry_looking, [-2, -2, -2, -1, -1, -1], [dict(r=0.05)])
        self.assertAllTrueOnAllArrayTypes(symmetry_looking, [-0.9, -0.900001], [dict(r=0.05)])

    def test_has_duplicate_max(self):
        self.assertTrueOnAllArrayTypes(has_duplicate_max, [2.1, 0, 0, 2.1, 1.1])
        self.assertFalseOnAllArrayTypes(has_duplicate_max, np.array([2.1, 0, 0, 2, 1.1]))
        self.assertTrueOnAllArrayTypes(has_duplicate_max, [1, 1, 1, 1])
        self.assertFalseOnAllArrayTypes(has_duplicate_max, np.array([0]))
        self.assertTrueOnAllArrayTypes(has_duplicate_max, np.array([1, 1]))

    def test_has_duplicate_min(self):
        self.assertTrueOnAllArrayTypes(has_duplicate_min, [-2.1, 0, 0, -2.1, 1.1])
        self.assertFalseOnAllArrayTypes(has_duplicate_min, [2.1, 0, -1, 2, 1.1])
        self.assertTrueOnAllArrayTypes(has_duplicate_min, np.array([1, 1, 1, 1]))
        self.assertFalseOnAllArrayTypes(has_duplicate_min, np.array([0]))
        self.assertTrueOnAllArrayTypes(has_duplicate_min, np.array([1, 1]))

    def test_has_duplicate(self):
        self.assertTrueOnAllArrayTypes(has_duplicate, np.array([-2.1, 0, 0, -2.1]))
        self.assertTrueOnAllArrayTypes(has_duplicate, [-2.1, 2.1, 2.1, 2.1])
        self.assertFalseOnAllArrayTypes(has_duplicate, [1.1, 1.2, 1.3, 1.4])
        self.assertFalseOnAllArrayTypes(has_duplicate, [1])
        self.assertFalseOnAllArrayTypes(has_duplicate, [])

    def test_sum(self):
        self.assertEqualOnAllArrayTypes(sum_values, [1, 2, 3, 4.1], 10.1)
        self.assertEqualOnAllArrayTypes(sum_values, [-1.2, -2, -3, -4], -10.2)
        self.assertEqualOnAllArrayTypes(sum_values, [], 0)

    def test_agg_autocorrelation_returns_correct_values(self):

        param = [{"f_agg": "mean", "maxlag": 10}]
        x = [1, 1, 1, 1, 1, 1, 1]
        expected_res = 0
        res = dict(agg_autocorrelation(x, param=param))["f_agg_\"mean\"__maxlag_10"]
        self.assertAlmostEqual(res, expected_res, places=4)

        x = [1, 2, -3]
        expected_res = 1 / np.var(x) * (((1 * 2 + 2 * (-3)) / 2 + (1 * -3)) / 2)
        res = dict(agg_autocorrelation(x, param=param))["f_agg_\"mean\"__maxlag_10"]
        self.assertAlmostEqual(res, expected_res, places=4)

        np.random.seed(42)
        x = np.random.normal(size=3000)
        expected_res = 0
        res = dict(agg_autocorrelation(x, param=param))["f_agg_\"mean\"__maxlag_10"]
        self.assertAlmostEqual(res, expected_res, places=2)

        param = [{"f_agg": "median", "maxlag": 10}]
        x = [1, 1, 1, 1, 1, 1, 1]
        expected_res = 0
        res = dict(agg_autocorrelation(x, param=param))["f_agg_\"median\"__maxlag_10"]
        self.assertAlmostEqual(res, expected_res, places=4)

        x = [1, 2, -3]
        expected_res = 1 / np.var(x) * (((1 * 2 + 2 * (-3)) / 2 + (1 * -3)) / 2)
        res = dict(agg_autocorrelation(x, param=param))["f_agg_\"median\"__maxlag_10"]
        self.assertAlmostEqual(res, expected_res, places=4)

    def test_agg_autocorrelation_returns_max_lag_does_not_affect_other_results(self):

        param = [{"f_agg": "mean", "maxlag": 1},
                 {"f_agg": "mean", "maxlag": 10}]
        x = range(10)
        res1 = dict(agg_autocorrelation(x, param=param))["f_agg_\"mean\"__maxlag_1"]
        res10 = dict(agg_autocorrelation(x, param=param))["f_agg_\"mean\"__maxlag_10"]
        self.assertAlmostEqual(res1, 0.77777777, places=4)
        self.assertAlmostEqual(res10, -0.64983164983165, places=4)

        param = [{"f_agg": "mean", "maxlag": 1}]
        x = range(10)
        res1 = dict(agg_autocorrelation(x, param=param))["f_agg_\"mean\"__maxlag_1"]
        self.assertAlmostEqual(res1, 0.77777777, places=4)

    def test_partial_autocorrelation(self):

        # Test for altering time series
        # len(x) < max_lag
        param = [{"lag": lag} for lag in range(10)]
        x = [1, 2, 1, 2, 1, 2]
        expected_res = [("lag_0", 1.0), ("lag_1", -1.0), ("lag_2", np.nan)]
        res = partial_autocorrelation(x, param=param)
        self.assertAlmostEqual(res[0][1], expected_res[0][1], places=4)
        self.assertAlmostEqual(res[1][1], expected_res[1][1], places=4)
        self.assertIsNaN(res[2][1])

        # Linear signal
        param = [{"lag": lag} for lag in range(10)]
        x = np.linspace(0, 1, 3000)
        expected_res = [("lag_0", 1.0), ("lag_1", 1.0), ("lag_2", 0)]
        res = partial_autocorrelation(x, param=param)
        self.assertAlmostEqual(res[0][1], expected_res[0][1], places=2)
        self.assertAlmostEqual(res[1][1], expected_res[1][1], places=2)
        self.assertAlmostEqual(res[2][1], expected_res[2][1], places=2)

        # Random noise
        np.random.seed(42)
        x = np.random.normal(size=3000)
        param = [{"lag": lag} for lag in range(10)]
        expected_res = [("lag_0", 1.0), ("lag_1", 0), ("lag_2", 0)]
        res = partial_autocorrelation(x, param=param)
        self.assertAlmostEqual(res[0][1], expected_res[0][1], places=1)
        self.assertAlmostEqual(res[1][1], expected_res[1][1], places=1)
        self.assertAlmostEqual(res[2][1], expected_res[2][1], places=1)

        # On a simulated AR process
        np.random.seed(42)
        param = [{"lag": lag} for lag in range(10)]
        # Simulate AR process
        T = 3000
        epsilon = np.random.randn(T)
        x = np.repeat(1.0, T)
        for t in range(T - 1):
            x[t + 1] = 0.5 * x[t] + 2 + epsilon[t]
        expected_res = [("lag_0", 1.0), ("lag_1", 0.5), ("lag_2", 0)]
        res = partial_autocorrelation(x, param=param)
        self.assertAlmostEqual(res[0][1], expected_res[0][1], places=1)
        self.assertAlmostEqual(res[1][1], expected_res[1][1], places=1)
        self.assertAlmostEqual(res[2][1], expected_res[2][1], places=1)

        # Some pathological cases
        param = [{"lag": lag} for lag in range(10)]
        # List of length 1
        res = partial_autocorrelation([1], param=param)
        for lag_no, lag_val in res:
            self.assertIsNaN(lag_val)
        # Empty list
        res = partial_autocorrelation([], param=param)
        for lag_no, lag_val in res:
            self.assertIsNaN(lag_val)
        # List contains only zeros
        res = partial_autocorrelation(np.zeros(100), param=param)
        for lag_no, lag_val in res:
            if lag_no == "lag_0":
                self.assertEqual(lag_val, 1.0)
            else:
                self.assertIsNaN(lag_val)

    def test_augmented_dickey_fuller(self):
        # todo: add unit test for the values of the test statistic

        # the adf hypothesis test checks for unit roots,
        # so H_0 = {random drift} vs H_1 = {AR(1) model}

        # H0 is true
        np.random.seed(seed=42)
        x = np.cumsum(np.random.uniform(size=100))
        param = [
            {"autolag": "BIC", "attr": "teststat"},
            {"autolag": "BIC", "attr": "pvalue"},
            {"autolag": "BIC", "attr": "usedlag"}
        ]
        expected_index = [
            'attr_"teststat"__autolag_"BIC"',
            'attr_"pvalue"__autolag_"BIC"',
            'attr_"usedlag"__autolag_"BIC"',
        ]

        res = augmented_dickey_fuller(x=x, param=param)
        res = pd.Series(dict(res))
        self.assertCountEqual(list(res.index), expected_index)
        self.assertGreater(res['attr_"pvalue"__autolag_"BIC"'], 0.10)
        self.assertEqual(res['attr_"usedlag"__autolag_"BIC"'], 0)

        # H0 should be rejected for AR(1) model with x_{t} = 1/2 x_{t-1} + e_{t}
        np.random.seed(seed=42)
        e = np.random.normal(0.1, 0.1, size=100)
        m = 50
        x = [0] * m
        x[0] = 100
        for i in range(1, m):
            x[i] = x[i - 1] * 0.5 + e[i]
        param = [
            {"autolag": "AIC", "attr": "teststat"},
            {"autolag": "AIC", "attr": "pvalue"},
            {"autolag": "AIC", "attr": "usedlag"}
        ]
        expected_index = [
            'attr_"teststat"__autolag_"AIC"',
            'attr_"pvalue"__autolag_"AIC"',
            'attr_"usedlag"__autolag_"AIC"',
        ]

        res = augmented_dickey_fuller(x=x, param=param)
        res = pd.Series(dict(res))
        self.assertCountEqual(list(res.index), expected_index)
        self.assertLessEqual(res['attr_"pvalue"__autolag_"AIC"'], 0.05)
        self.assertEqual(res['attr_"usedlag"__autolag_"AIC"'], 0)

        # Check if LinAlgError and ValueError are catched
        res_linalg_error = augmented_dickey_fuller(x=np.repeat(np.nan, 100), param=param)
        res_value_error = augmented_dickey_fuller(x=[], param=param)
        for index, val in res_linalg_error:
            self.assertIsNaN(val)
        for index, val in res_value_error:
            self.assertIsNaN(val)

        # Should return NaN if "attr" is unknown
        res_attr_error = augmented_dickey_fuller(x=x, param=[{"autolag": "AIC", "attr": ""}])
        for index, val in res_attr_error:
            self.assertIsNaN(val)

    def test_abs_energy(self):
        self.assertEqualOnAllArrayTypes(abs_energy, [1, 1, 1], 3)
        self.assertEqualOnAllArrayTypes(abs_energy, [1, 2, 3], 14)
        self.assertEqualOnAllArrayTypes(abs_energy, [-1, 2, -3], 14)
        self.assertAlmostEqualOnAllArrayTypes(abs_energy, [-1, 1.3], 2.69)
        self.assertEqualOnAllArrayTypes(abs_energy, [1], 1)

    def test_cid_ce(self):
        self.assertEqualOnAllArrayTypes(cid_ce, [1, 1, 1], 0, normalize=True)
        self.assertEqualOnAllArrayTypes(cid_ce, [0, 4], 2, normalize=True)
        self.assertEqualOnAllArrayTypes(cid_ce, [100, 104], 2, normalize=True)

        self.assertEqualOnAllArrayTypes(cid_ce, [1, 1, 1], 0, normalize=False)
        self.assertEqualOnAllArrayTypes(cid_ce, [0.5, 3.5, 7.5], 5, normalize=False)
        self.assertEqualOnAllArrayTypes(cid_ce, [-4.33, -1.33, 2.67], 5, normalize=False)

    def test_lempel_ziv_complexity(self):
        self.assertAlmostEqualOnAllArrayTypes(lempel_ziv_complexity, [1, 1, 1], 2. / 3, bins=2)
        self.assertAlmostEqualOnAllArrayTypes(lempel_ziv_complexity, [1, 1, 1], 2. / 3, bins=5)

        self.assertAlmostEqualOnAllArrayTypes(lempel_ziv_complexity, [1, 1, 1, 1, 1, 1, 1],
                                              0.4285714285, bins=2)
        self.assertAlmostEqualOnAllArrayTypes(lempel_ziv_complexity, [1, 1, 1, 2, 1, 1, 1],
                                              0.5714285714, bins=2)

        self.assertAlmostEqualOnAllArrayTypes(lempel_ziv_complexity,
                                              [-1, 4.3, 5, 1, -4.5, 1, 5, 7, -3.4, 6],
                                              0.8, bins=10)
        self.assertAlmostEqualOnAllArrayTypes(lempel_ziv_complexity,
                                              [-1, np.nan, 5, 1, -4.5, 1, 5, 7, -3.4, 6],
                                              0.4, bins=10)

    def test_fourier_entropy(self):
        self.assertAlmostEqualOnAllArrayTypes(fourier_entropy, [1, 2, 1], 0.693147180, bins=2)
        self.assertAlmostEqualOnAllArrayTypes(fourier_entropy, [1, 2, 1], 0.693147180, bins=5)

        self.assertAlmostEqualOnAllArrayTypes(fourier_entropy, [1, 1, 2, 1, 1, 1, 1],
                                              0.5623351446188083, bins=5)
        self.assertAlmostEqualOnAllArrayTypes(fourier_entropy, [1, 1, 1, 1, 2, 1, 1],
                                              1.0397207708399179, bins=5)

        self.assertAlmostEqualOnAllArrayTypes(fourier_entropy,
                                              [-1, 4.3, 5, 1, -4.5, 1, 5, 7, -3.4, 6],
                                              1.5607104090414063, bins=10)
        self.assertIsNanOnAllArrayTypes(fourier_entropy,
                                        [-1, np.nan, 5, 1, -4.5, 1, 5, 7, -3.4, 6],
                                        bins=10)

    def test_permutation_entropy(self):
        self.assertAlmostEqualOnAllArrayTypes(permutation_entropy, [4, 7, 9, 10, 6, 11, 3], 1.054920167,
                                              dimension=3, tau=1)

        # should grow
        self.assertAlmostEqualOnAllArrayTypes(permutation_entropy, [1, -1, 1, -1, 1, -1, 1, -1],
                                              0.6931471805599453, dimension=3, tau=1)
        self.assertAlmostEqualOnAllArrayTypes(permutation_entropy, [1, -1, 1, -1, 1,  1, 1, -1],
                                              1.3296613488547582, dimension=3, tau=1)

        self.assertAlmostEqualOnAllArrayTypes(permutation_entropy,
                                              [-1, 4.3, 5, 1, -4.5, 1, 5, 7, -3.4, 6],
                                              1.0397207708399179, dimension=3, tau=2)
        # nan is treated like any other number
        self.assertAlmostEqualOnAllArrayTypes(permutation_entropy,
                                              [-1, 4.3, 5, 1, -4.5, 1, 5, np.nan, -3.4, 6],
                                              1.0397207708399179, dimension=3, tau=2)

        # if too short, return nan
        self.assertIsNanOnAllArrayTypes(permutation_entropy, [1, -1], dimension=3, tau=1)

    def test_ratio_beyond_r_sigma(self):

        x = [0, 1] * 10 + [10, 20, -30]  # std of x is 7.21, mean 3.04
        self.assertEqualOnAllArrayTypes(ratio_beyond_r_sigma, x, 3. / len(x), r=1)
        self.assertEqualOnAllArrayTypes(ratio_beyond_r_sigma, x, 2. / len(x), r=2)
        self.assertEqualOnAllArrayTypes(ratio_beyond_r_sigma, x, 1. / len(x), r=3)
        self.assertEqualOnAllArrayTypes(ratio_beyond_r_sigma, x, 0, r=20)

    def test_mean_abs_change(self):
        self.assertEqualOnAllArrayTypes(mean_abs_change, [-2, 2, 5], 3.5)
        self.assertEqualOnAllArrayTypes(mean_abs_change, [1, 2, -1], 2)

    def test_mean_change(self):
        self.assertEqualOnAllArrayTypes(mean_change, [-2, 2, 5], 3.5)
        self.assertEqualOnAllArrayTypes(mean_change, [1, 2, -1], -1)
        self.assertEqualOnAllArrayTypes(mean_change, [10, 20], 10)
        self.assertIsNanOnAllArrayTypes(mean_change, [1])
        self.assertIsNanOnAllArrayTypes(mean_change, [])

    def test_mean_second_derivate_central(self):
        self.assertEqualOnAllArrayTypes(mean_second_derivative_central, list(range(10)), 0)
        self.assertEqualOnAllArrayTypes(mean_second_derivative_central, [1, 3, 5], 0)
        self.assertEqualOnAllArrayTypes(mean_second_derivative_central, [1, 3, 7, -3], -3)

    def test_median(self):
        self.assertEqualOnAllArrayTypes(median, [1, 1, 2, 2], 1.5)
        self.assertEqualOnAllArrayTypes(median, [0.5, 0.5, 2, 3.5, 10], 2)
        self.assertEqualOnAllArrayTypes(median, [0.5], 0.5)
        self.assertIsNanOnAllArrayTypes(median, [])

    def test_mean(self):
        self.assertEqualOnAllArrayTypes(mean, [1, 1, 2, 2], 1.5)
        self.assertEqualOnAllArrayTypes(mean, [0.5, 0.5, 2, 3.5, 10], 3.3)
        self.assertEqualOnAllArrayTypes(mean, [0.5], 0.5)
        self.assertIsNanOnAllArrayTypes(mean, [])

    def test_length(self):
        self.assertEqualOnAllArrayTypes(length, [1, 2, 3, 4], 4)
        self.assertEqualOnAllArrayTypes(length, [1, 2, 3], 3)
        self.assertEqualOnAllArrayTypes(length, [1, 2], 2)
        self.assertEqualOnAllArrayTypes(length, [1, 2, 3, np.NaN], 4)
        self.assertEqualOnAllArrayTypes(length, [], 0)

    def test_standard_deviation(self):
        self.assertAlmostEqualOnAllArrayTypes(standard_deviation, [1, 1, -1, -1], 1)
        self.assertAlmostEqualOnAllArrayTypes(standard_deviation, [1, 2, -2, -1], 1.58113883008)
        self.assertIsNanOnAllArrayTypes(standard_deviation, [])

    def test_variation_coefficient(self):
        self.assertIsNanOnAllArrayTypes(variation_coefficient, [1, 1, -1, -1], )
        self.assertAlmostEqualOnAllArrayTypes(variation_coefficient, [1, 2, -3, -1], -7.681145747868608)
        self.assertAlmostEqualOnAllArrayTypes(variation_coefficient, [1, 2, 4, -1], 1.2018504251546631)
        self.assertIsNanOnAllArrayTypes(variation_coefficient, [])

    def test_variance(self):
        self.assertAlmostEqualOnAllArrayTypes(variance, [1, 1, -1, -1], 1)
        self.assertAlmostEqualOnAllArrayTypes(variance, [1, 2, -2, -1], 2.5)
        self.assertIsNanOnAllArrayTypes(variance, [])

    def test_skewness(self):
        self.assertEqualOnAllArrayTypes(skewness, [1, 1, 1, 2, 2, 2], 0)
        self.assertAlmostEqualOnAllArrayTypes(skewness, [1, 1, 1, 2, 2], 0.6085806194501855)
        self.assertEqualOnAllArrayTypes(skewness, [1, 1, 1], 0)
        self.assertIsNanOnAllArrayTypes(skewness, [1, 1])

    def test_kurtosis(self):
        self.assertAlmostEqualOnAllArrayTypes(kurtosis, [1, 1, 1, 2, 2], -3.333333333333333)
        self.assertAlmostEqualOnAllArrayTypes(kurtosis, [1, 1, 1, 1], 0)
        self.assertIsNanOnAllArrayTypes(kurtosis, [1, 1, 1])

    def test_absolute_sum_of_changes(self):
        self.assertEqualOnAllArrayTypes(absolute_sum_of_changes, [1, 1, 1, 1, 2, 1], 2)
        self.assertEqualOnAllArrayTypes(absolute_sum_of_changes, [1, -1, 1, -1], 6)
        self.assertEqualOnAllArrayTypes(absolute_sum_of_changes, [1], 0)
        self.assertEqualOnAllArrayTypes(absolute_sum_of_changes, [], 0)

    def test_longest_strike_below_mean(self):
        self.assertEqualOnAllArrayTypes(longest_strike_below_mean, [1, 2, 1, 1, 1, 2, 2, 2], 3)
        self.assertEqualOnAllArrayTypes(longest_strike_below_mean, [1, 2, 3, 4, 5, 6], 3)
        self.assertEqualOnAllArrayTypes(longest_strike_below_mean, [1, 2, 3, 4, 5], 2)
        self.assertEqualOnAllArrayTypes(longest_strike_below_mean, [1, 2, 1], 1)
        self.assertEqualOnAllArrayTypes(longest_strike_below_mean, [], 0)

    def test_longest_strike_above_mean(self):
        self.assertEqualOnAllArrayTypes(longest_strike_above_mean, [1, 2, 1, 2, 1, 2, 2, 1], 2)
        self.assertEqualOnAllArrayTypes(longest_strike_above_mean, [1, 2, 3, 4, 5, 6], 3)
        self.assertEqualOnAllArrayTypes(longest_strike_above_mean, [1, 2, 3, 4, 5], 2)
        self.assertEqualOnAllArrayTypes(longest_strike_above_mean, [1, 2, 1], 1)
        self.assertEqualOnAllArrayTypes(longest_strike_above_mean, [], 0)

    def test_count_above_mean(self):
        self.assertEqualOnAllArrayTypes(count_above_mean, [1, 2, 1, 2, 1, 2], 3)
        self.assertEqualOnAllArrayTypes(count_above_mean, [1, 1, 1, 1, 1, 2], 1)
        self.assertEqualOnAllArrayTypes(count_above_mean, [1, 1, 1, 1, 1], 0)
        self.assertEqualOnAllArrayTypes(count_above_mean, [], 0)

    def test_count_below_mean(self):
        self.assertEqualOnAllArrayTypes(count_below_mean, [1, 2, 1, 2, 1, 2], 3)
        self.assertEqualOnAllArrayTypes(count_below_mean, [1, 1, 1, 1, 1, 2], 5)
        self.assertEqualOnAllArrayTypes(count_below_mean, [1, 1, 1, 1, 1], 0)
        self.assertEqualOnAllArrayTypes(count_below_mean, [], 0)

    def test_last_location_maximum(self):
        self.assertAlmostEqualOnAllArrayTypes(last_location_of_maximum, [1, 2, 1, 2, 1], 0.8)
        self.assertAlmostEqualOnAllArrayTypes(last_location_of_maximum, [1, 2, 1, 1, 2], 1.0)
        self.assertAlmostEqualOnAllArrayTypes(last_location_of_maximum, [2, 1, 1, 1, 1], 0.2)
        self.assertAlmostEqualOnAllArrayTypes(last_location_of_maximum, [1, 1, 1, 1, 1], 1.0)
        self.assertAlmostEqualOnAllArrayTypes(last_location_of_maximum, [1], 1.0)
        self.assertIsNanOnAllArrayTypes(last_location_of_maximum, [])

    def test_first_location_of_maximum(self):
        self.assertAlmostEqualOnAllArrayTypes(first_location_of_maximum, [1, 2, 1, 2, 1], 0.2)
        self.assertAlmostEqualOnAllArrayTypes(first_location_of_maximum, [1, 2, 1, 1, 2], 0.2)
        self.assertAlmostEqualOnAllArrayTypes(first_location_of_maximum, [2, 1, 1, 1, 1], 0.0)
        self.assertAlmostEqualOnAllArrayTypes(first_location_of_maximum, [1, 1, 1, 1, 1], 0.0)
        self.assertAlmostEqualOnAllArrayTypes(first_location_of_maximum, [1], 0.0)
        self.assertIsNanOnAllArrayTypes(first_location_of_maximum, [])

    def test_last_location_of_minimum(self):
        self.assertAlmostEqualOnAllArrayTypes(last_location_of_minimum, [1, 2, 1, 2, 1], 1.0)
        self.assertAlmostEqualOnAllArrayTypes(last_location_of_minimum, [1, 2, 1, 2, 2], 0.6)
        self.assertAlmostEqualOnAllArrayTypes(last_location_of_minimum, [2, 1, 1, 1, 2], 0.8)
        self.assertAlmostEqualOnAllArrayTypes(last_location_of_minimum, [1, 1, 1, 1, 1], 1.0)
        self.assertAlmostEqualOnAllArrayTypes(last_location_of_minimum, [1], 1.0)
        self.assertIsNanOnAllArrayTypes(last_location_of_minimum, [])

    def test_first_location_of_minimum(self):
        self.assertAlmostEqualOnAllArrayTypes(first_location_of_minimum, [1, 2, 1, 2, 1], 0.0)
        self.assertAlmostEqualOnAllArrayTypes(first_location_of_minimum, [2, 2, 1, 2, 2], 0.4)
        self.assertAlmostEqualOnAllArrayTypes(first_location_of_minimum, [2, 1, 1, 1, 2], 0.2)
        self.assertAlmostEqualOnAllArrayTypes(first_location_of_minimum, [1, 1, 1, 1, 1], 0.0)
        self.assertAlmostEqualOnAllArrayTypes(first_location_of_minimum, [1], 0.0)
        self.assertIsNanOnAllArrayTypes(first_location_of_minimum, [])

    def test_percentage_of_doubled_datapoints(self):
        self.assertAlmostEqualOnAllArrayTypes(percentage_of_reoccurring_datapoints_to_all_datapoints, [1, 1, 2, 3, 4],
                                              0.4)
        self.assertAlmostEqualOnAllArrayTypes(percentage_of_reoccurring_datapoints_to_all_datapoints, [1, 1.5, 2, 3], 0)
        self.assertAlmostEqualOnAllArrayTypes(percentage_of_reoccurring_datapoints_to_all_datapoints, [1], 0)
        self.assertAlmostEqualOnAllArrayTypes(percentage_of_reoccurring_datapoints_to_all_datapoints,
                                              [1.111, -2.45, 1.111, 2.45], 0.5)
        self.assertIsNanOnAllArrayTypes(percentage_of_reoccurring_datapoints_to_all_datapoints, [])

    def test_ratio_of_doubled_values(self):
        self.assertAlmostEqualOnAllArrayTypes(percentage_of_reoccurring_values_to_all_values, [1, 1, 2, 3, 4], 0.25)
        self.assertAlmostEqualOnAllArrayTypes(percentage_of_reoccurring_values_to_all_values, [1, 1.5, 2, 3], 0)
        self.assertAlmostEqualOnAllArrayTypes(percentage_of_reoccurring_values_to_all_values, [1], 0)
        self.assertAlmostEqualOnAllArrayTypes(percentage_of_reoccurring_values_to_all_values,
                                              [1.111, -2.45, 1.111, 2.45], 1.0 / 3.0)
        self.assertIsNanOnAllArrayTypes(percentage_of_reoccurring_values_to_all_values, [])

    def test_sum_of_reoccurring_values(self):
        self.assertAlmostEqualOnAllArrayTypes(sum_of_reoccurring_values, [1, 1, 2, 3, 4, 4], 5)
        self.assertAlmostEqualOnAllArrayTypes(sum_of_reoccurring_values, [1, 1.5, 2, 3], 0)
        self.assertAlmostEqualOnAllArrayTypes(sum_of_reoccurring_values, [1], 0)
        self.assertAlmostEqualOnAllArrayTypes(sum_of_reoccurring_values, [1.111, -2.45, 1.111, 2.45], 1.111)
        self.assertAlmostEqualOnAllArrayTypes(sum_of_reoccurring_values, [], 0)

    def test_sum_of_reoccurring_data_points(self):
        self.assertAlmostEqualOnAllArrayTypes(sum_of_reoccurring_data_points, [1, 1, 2, 3, 4, 4], 10)
        self.assertAlmostEqualOnAllArrayTypes(sum_of_reoccurring_data_points, [1, 1.5, 2, 3], 0)
        self.assertAlmostEqualOnAllArrayTypes(sum_of_reoccurring_data_points, [1], 0)
        self.assertAlmostEqualOnAllArrayTypes(sum_of_reoccurring_data_points, [1.111, -2.45, 1.111, 2.45], 2.222)
        self.assertAlmostEqualOnAllArrayTypes(sum_of_reoccurring_data_points, [], 0)

    def test_uniqueness_factor(self):
        self.assertAlmostEqualOnAllArrayTypes(ratio_value_number_to_time_series_length, [1, 1, 2, 3, 4], 0.8)
        self.assertAlmostEqualOnAllArrayTypes(ratio_value_number_to_time_series_length, [1, 1.5, 2, 3], 1)
        self.assertAlmostEqualOnAllArrayTypes(ratio_value_number_to_time_series_length, [1], 1)
        self.assertAlmostEqualOnAllArrayTypes(ratio_value_number_to_time_series_length, [1.111, -2.45, 1.111, 2.45],
                                              0.75)
        self.assertIsNanOnAllArrayTypes(ratio_value_number_to_time_series_length, [])

    def test_fft_coefficient(self):
        x = range(10)
        param = [{"coeff": 0, "attr": "real"}, {"coeff": 1, "attr": "real"}, {"coeff": 2, "attr": "real"},
                 {"coeff": 0, "attr": "imag"}, {"coeff": 1, "attr": "imag"}, {"coeff": 2, "attr": "imag"},
                 {"coeff": 0, "attr": "angle"}, {"coeff": 1, "attr": "angle"}, {"coeff": 2, "attr": "angle"},
                 {"coeff": 0, "attr": "abs"}, {"coeff": 1, "attr": "abs"}, {"coeff": 2, "attr": "abs"}]
        expected_index = ['attr_"real"__coeff_0', 'attr_"real"__coeff_1', 'attr_"real"__coeff_2',
                          'attr_"imag"__coeff_0', 'attr_"imag"__coeff_1', 'attr_"imag"__coeff_2',
                          'attr_"angle"__coeff_0', 'attr_"angle"__coeff_1', 'attr_"angle"__coeff_2',
                          'attr_"abs"__coeff_0', 'attr_"abs"__coeff_1', 'attr_"abs"__coeff_2']

        res = pd.Series(dict(fft_coefficient(x, param)))
        self.assertCountEqual(list(res.index), expected_index)
        self.assertAlmostEqual(res['attr_"imag"__coeff_0'], 0, places=6)
        self.assertAlmostEqual(res['attr_"real"__coeff_0'], sum(x), places=6)
        self.assertAlmostEqual(res['attr_"angle"__coeff_0'], 0, places=6)
        self.assertAlmostEqual(res['attr_"abs"__coeff_0'], sum(x), places=6)

        x = [0, 1, 0, 0]
        res = pd.Series(dict(fft_coefficient(x, param)))
        # see documentation of fft in numpy
        # should return array([1. + 0.j, 0. - 1.j, -1. + 0.j])
        self.assertAlmostEqual(res['attr_"imag"__coeff_0'], 0, places=6)
        self.assertAlmostEqual(res['attr_"real"__coeff_0'], 1, places=6)
        self.assertAlmostEqual(res['attr_"imag"__coeff_1'], -1, places=6)
        self.assertAlmostEqual(res['attr_"angle"__coeff_1'], -90, places=6)
        self.assertAlmostEqual(res['attr_"real"__coeff_1'], 0, places=6)
        self.assertAlmostEqual(res['attr_"imag"__coeff_2'], 0, places=6)
        self.assertAlmostEqual(res['attr_"real"__coeff_2'], -1, places=6)

        # test what happens if coeff is biger than time series lenght
        x = range(5)
        param = [{"coeff": 10, "attr": "real"}]
        expected_index = ['attr_"real"__coeff_10']

        res = pd.Series(dict(fft_coefficient(x, param)))
        self.assertCountEqual(list(res.index), expected_index)
        self.assertIsNaN(res['attr_"real"__coeff_10'])

    def test_fft_aggregated(self):
        param = [
            {"aggtype": "centroid"},
            {"aggtype": "variance"},
            {"aggtype": "skew"},
            {"aggtype": "kurtosis"}
        ]
        expected_index = ['aggtype_"centroid"', 'aggtype_"variance"', 'aggtype_"skew"', 'aggtype_"kurtosis"']

        x = np.arange(10)
        res = pd.Series(dict(fft_aggregated(x, param)))
        self.assertCountEqual(list(res.index), expected_index)
        self.assertAlmostEqual(res['aggtype_"centroid"'], 1.135, places=3)
        self.assertAlmostEqual(res['aggtype_"variance"'], 2.368, places=3)
        self.assertAlmostEqual(res['aggtype_"skew"'], 1.249, places=3)
        self.assertAlmostEqual(res['aggtype_"kurtosis"'], 3.643, places=3)

        # Scalar multiplying the distribution should not change the results:
        x = 10 * x
        res = pd.Series(dict(fft_aggregated(x, param)))
        self.assertCountEqual(list(res.index), expected_index)
        self.assertAlmostEqual(res['aggtype_"centroid"'], 1.135, places=3)
        self.assertAlmostEqual(res['aggtype_"variance"'], 2.368, places=3)
        self.assertAlmostEqual(res['aggtype_"skew"'], 1.249, places=3)
        self.assertAlmostEqual(res['aggtype_"kurtosis"'], 3.643, places=3)

        # The fft of a sign wave is a dirac delta, variance and skew should be near zero, kurtosis should be near 3:
        # However, in the discrete limit, skew and kurtosis blow up in a manner that is noise dependent and are
        # therefore bad features, therefore an nan should be returned for these values
        x = np.sin(2 * np.pi / 10 * np.arange(30))
        res = pd.Series(dict(fft_aggregated(x, param)))
        self.assertCountEqual(list(res.index), expected_index)
        self.assertAlmostEqual(res['aggtype_"centroid"'], 3., places=5)
        self.assertAlmostEqual(res['aggtype_"variance"'], 0., places=5)
        self.assertIsNaN(res['aggtype_"skew"'])
        self.assertIsNaN(res['aggtype_"kurtosis"'])

        # Gaussian test:
        def normal(y, mean_, sigma_):
            return 1 / (2 * np.pi * sigma_ ** 2) * np.exp(-(y - mean_) ** 2 / (2 * sigma_ ** 2))

        mean_ = 500.
        sigma_ = 1.
        range_ = int(2 * mean_)
        x = list(map(lambda x: normal(x, mean_, sigma_), range(range_)))

        # The fourier transform of a Normal dist in the positive halfspace is a half normal,
        # Hand calculated values of centroid and variance based for the half-normal dist:
        # (Ref: https://en.wikipedia.org/wiki/Half-normal_distribution)
        expected_fft_centroid = (range_ / (2 * np.pi * sigma_)) * np.sqrt(2 / np.pi)
        expected_fft_var = (range_ / (2 * np.pi * sigma_)) ** 2 * (1 - 2 / np.pi)

        # Calculate values for unit test:
        res = pd.Series(dict(fft_aggregated(x, param)))
        self.assertCountEqual(list(res.index), expected_index)

        # Compare against hand calculated values:
        rel_diff_allowed = 0.02
        self.assertAlmostEqual(
            res['aggtype_"centroid"'], expected_fft_centroid,
            delta=rel_diff_allowed * expected_fft_centroid
        )
        self.assertAlmostEqual(
            res['aggtype_"variance"'], expected_fft_var,
            delta=rel_diff_allowed * expected_fft_var
        )

    def test_number_peaks(self):
        x = np.array([0, 1, 2, 1, 0, 1, 2, 3, 4, 5, 4, 3, 2, 1])
        self.assertEqualOnAllArrayTypes(number_peaks, x, 2, 1)
        self.assertEqualOnAllArrayTypes(number_peaks, x, 2, 2)
        self.assertEqualOnAllArrayTypes(number_peaks, x, 1, 3)
        self.assertEqualOnAllArrayTypes(number_peaks, x, 1, 4)
        self.assertEqualOnAllArrayTypes(number_peaks, x, 0, 5)
        self.assertEqualOnAllArrayTypes(number_peaks, x, 0, 6)

    def test_mass_quantile(self):

        x = [1] * 101
        param = [{"q": 0.5}]
        expected_index = ["q_0.5"]
        res = index_mass_quantile(x, param)

        res = pd.Series(dict(res))
        self.assertCountEqual(list(res.index), expected_index)
        self.assertAlmostEqual(res["q_0.5"], 0.5, places=1)

        # Test for parts of pandas series
        x = pd.Series([0] * 55 + [1] * 101)
        param = [{"q": 0.5}]
        expected_index = ["q_0.5"]
        res = index_mass_quantile(x[x > 0], param)

        res = pd.Series(dict(res))
        self.assertCountEqual(list(res.index), expected_index)
        self.assertAlmostEqual(res["q_0.5"], 0.5, places=1)

        x = [0] * 1000 + [1]
        param = [{"q": 0.5}, {"q": 0.99}]
        expected_index = ["q_0.5", "q_0.99"]
        res = index_mass_quantile(x, param)

        res = pd.Series(dict(res))
        self.assertCountEqual(list(res.index), expected_index)
        self.assertAlmostEqual(res["q_0.5"], 1, places=1)
        self.assertAlmostEqual(res["q_0.99"], 1, places=1)

        x = [0, 1, 1, 0, 0, 1, 0, 0]
        param = [{"q": 0.30}, {"q": 0.60}, {"q": 0.90}]
        expected_index = ["q_0.3", "q_0.6",
                          "q_0.9"]
        res = index_mass_quantile(x, param)

        res = pd.Series(dict(res))

        self.assertCountEqual(list(res.index), expected_index)
        self.assertAlmostEqual(res["q_0.3"], 0.25, places=1)
        self.assertAlmostEqual(res["q_0.6"], 0.375, places=1)
        self.assertAlmostEqual(res["q_0.9"], 0.75, places=1)

        x = [0, 0, 0]
        param = [{"q": 0.5}]
        expected_index = ["q_0.5"]
        res = index_mass_quantile(x, param)

        res = pd.Series(dict(res))
        self.assertCountEqual(list(res.index), expected_index)
        self.assertTrue(np.isnan(res["q_0.5"]))

        x = []
        param = [{"q": 0.5}]
        expected_index = ["q_0.5"]
        res = index_mass_quantile(x, param)

        res = pd.Series(dict(res))
        self.assertCountEqual(list(res.index), expected_index)
        self.assertTrue(np.isnan(res["q_0.5"]))

    def test_number_cwt_peaks(self):
        x = [1, 1, 1, 1, 1, 1, 1, 5, 1, 1, 1, 1, 1, 1, 5, 1, 1, 1, 1, 1, 1]
        self.assertEqualOnAllArrayTypes(number_cwt_peaks, x, 2, 2)

    def test_spkt_welch_density(self):

        # todo: improve tests
        x = range(10)
        param = [{"coeff": 1}, {"coeff": 10}]
        expected_index = ["coeff_1", "coeff_10"]
        res = pd.Series(dict(spkt_welch_density(x, param)))
        self.assertCountEqual(list(res.index), expected_index)
        self.assertIsNaN(res["coeff_10"])

    def test_cwt_coefficients(self):
        x = [0.1, 0.2, 0.3]
        param = [{"widths": (1, 2, 3), "coeff": 2, "w": 1},
                 {"widths": (1, 3), "coeff": 2, "w": 3},
                 {"widths": (1, 3), "coeff": 5, "w": 3}]
        shuffle(param)

        expected_index = ["coeff_2__w_1__widths_(1, 2, 3)",
                          "coeff_2__w_3__widths_(1, 3)",
                          "coeff_5__w_3__widths_(1, 3)"]

        res = cwt_coefficients(x, param)
        res = pd.Series(dict(res))

        # todo: add unit test for the values
        self.assertCountEqual(list(res.index), expected_index)
        self.assertTrue(math.isnan(res["coeff_5__w_3__widths_(1, 3)"]))

    def test_ar_coefficient(self):

        # Test for X_i = 2.5 * X_{i-1} + 1
        param = [{"k": 1, "coeff": 0}, {"k": 1, "coeff": 1}]
        shuffle(param)

        x = [1] + 9 * [0]
        for i in range(1, len(x)):
            x[i] = 2.5 * x[i - 1] + 1

        res = ar_coefficient(x, param)
        expected_index = ["coeff_0__k_1", "coeff_1__k_1"]

        res = pd.Series(dict(res))
        self.assertCountEqual(list(res.index), expected_index)
        self.assertAlmostEqual(res["coeff_0__k_1"], 1, places=2)
        self.assertAlmostEqual(res["coeff_1__k_1"], 2.5, places=2)

        # Test for X_i = 1.4 * X_{i-1} - 1 X_{i-2} + 1
        param = [{"k": 1, "coeff": 0}, {"k": 1, "coeff": 1},
                 {"k": 2, "coeff": 0}, {"k": 2, "coeff": 1}, {"k": 2, "coeff": 2}, {"k": 2, "coeff": 3}]
        shuffle(param)

        x = [1, 1] + 5 * [0]
        for i in range(2, len(x)):
            x[i] = (-2) * x[i - 2] + 3.5 * x[i - 1] + 1

        res = ar_coefficient(x, param)
        expected_index = ["coeff_0__k_1", "coeff_1__k_1",
                          "coeff_0__k_2", "coeff_1__k_2",
                          "coeff_2__k_2", "coeff_3__k_2"]

        res = pd.Series(dict(res))

        self.assertIsInstance(res, pd.Series)
        self.assertCountEqual(list(res.index), expected_index)
        self.assertAlmostEqual(res["coeff_0__k_2"], 1, places=2)
        self.assertAlmostEqual(res["coeff_1__k_2"], 3.5, places=2)
        self.assertAlmostEqual(res["coeff_2__k_2"], -2, places=2)
        self.assertTrue(np.isnan(res["coeff_3__k_2"]))

    def test_time_reversal_asymmetry_statistic(self):
        x = [1] * 10
        self.assertAlmostEqualOnAllArrayTypes(time_reversal_asymmetry_statistic, x, 0, 0)
        self.assertAlmostEqualOnAllArrayTypes(time_reversal_asymmetry_statistic, x, 0, 1)
        self.assertAlmostEqualOnAllArrayTypes(time_reversal_asymmetry_statistic, x, 0, 2)
        self.assertAlmostEqualOnAllArrayTypes(time_reversal_asymmetry_statistic, x, 0, 3)

        x = [1, 2, -3, 4]
        # 1/2 * ( (4^2 * -3 + 3 * 2^2) + (3^2*2)-(2*1^1)) = 1/2 * (-48+12+18-2) = 20/2
        self.assertAlmostEqualOnAllArrayTypes(time_reversal_asymmetry_statistic, x, -10, 1)
        self.assertAlmostEqualOnAllArrayTypes(time_reversal_asymmetry_statistic, x, 0, 2)
        self.assertAlmostEqualOnAllArrayTypes(time_reversal_asymmetry_statistic, x, 0, 3)

    def test_number_crossing_m(self):
        x = [10, -10, 10, -10]
        self.assertEqualOnAllArrayTypes(number_crossing_m, x, 3, 0)
        self.assertEqualOnAllArrayTypes(number_crossing_m, x, 0, 10)

        x = [10, 20, 20, 30]
        self.assertEqualOnAllArrayTypes(number_crossing_m, x, 0, 0)
        self.assertEqualOnAllArrayTypes(number_crossing_m, x, 1, 15)

    def test_c3(self):
        x = [1] * 10
        self.assertAlmostEqualOnAllArrayTypes(c3, x, 1, 0)
        self.assertAlmostEqualOnAllArrayTypes(c3, x, 1, 1)
        self.assertAlmostEqualOnAllArrayTypes(c3, x, 1, 2)
        self.assertAlmostEqualOnAllArrayTypes(c3, x, 1, 3)

        x = [1, 2, -3, 4]
        # 1/2 *(1*2*(-3)+2*(-3)*4) = 1/2 *(-6-24) = -30/2
        self.assertAlmostEqualOnAllArrayTypes(c3, x, -15, 1)
        self.assertAlmostEqualOnAllArrayTypes(c3, x, 0, 2)
        self.assertAlmostEqualOnAllArrayTypes(c3, x, 0, 3)

    def test_binned_entropy(self):
        self.assertAlmostEqualOnAllArrayTypes(binned_entropy, [10] * 100, 0, 10)
        self.assertAlmostEqualOnAllArrayTypes(binned_entropy, [10] * 10 + [1], - (10 / 11 * np.math.log(10 / 11) +
                                                                                  1 / 11 * np.math.log(1 / 11)), 10)
        self.assertAlmostEqualOnAllArrayTypes(binned_entropy, [10] * 10 + [1], - (10 / 11 * np.math.log(10 / 11) +
                                                                                  1 / 11 * np.math.log(1 / 11)), 10)
        self.assertAlmostEqualOnAllArrayTypes(binned_entropy, [10] * 10 + [1], - (10 / 11 * np.math.log(10 / 11) +
                                                                                  1 / 11 * np.math.log(1 / 11)), 100)
        self.assertAlmostEqualOnAllArrayTypes(binned_entropy, list(range(10)), - np.math.log(1 / 10), 100)
        self.assertAlmostEqualOnAllArrayTypes(binned_entropy, list(range(100)), - np.math.log(1 / 2), 2)

    def test_sample_entropy(self):
        # "random" list -> large entropy
        ts = [1, 4, 5, 1, 7, 3, 1, 2, 5, 8, 9, 7, 3, 7, 9, 5, 4, 3, 9, 1, 2, 3, 4, 2, 9, 6, 7, 4, 9, 2, 9, 9, 6, 5, 1,
              3, 8, 1, 5, 3, 8, 4, 1, 2, 2, 1, 6, 5, 3, 6, 5, 4, 8, 9, 6, 7, 5, 3, 2, 5, 4, 2, 5, 1, 6, 5, 3, 5, 6, 7,
              8, 5, 2, 8, 6, 3, 8, 2, 7, 1, 7, 3, 5, 6, 2, 1, 3, 7, 3, 5, 3, 7, 6, 7, 7, 2, 3, 1, 7, 8]
        self.assertAlmostEqualOnAllArrayTypes(sample_entropy, ts, 2.38262780)
        # This is not very complex, so it gives a small value
        ts = [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
        self.assertAlmostEqualOnAllArrayTypes(sample_entropy, ts, 0.25131442)
        # however adding a 2 increases complexity
        ts = [1, 1, 2, 1, 1, 1, 1, 1, 1, 1]
        self.assertAlmostEqualOnAllArrayTypes(sample_entropy, ts, 0.74193734)
        # and it does not matter where
        ts = [1, 1, 1, 2, 1, 1, 1, 1, 1, 1]
        self.assertAlmostEqualOnAllArrayTypes(sample_entropy, ts, 0.74193734)
        # negative numbers also work
        ts = [1, -1, 1, -1, 1, -1]
        self.assertAlmostEqualOnAllArrayTypes(sample_entropy, ts, 0.69314718)
        # nan gives nan
        ts = [1, -1, 1, np.nan, 1, -1]
        self.assertIsNanOnAllArrayTypes(sample_entropy, ts)
        # this is not a very "random" list, so it should give a small entropy
        ts = list(range(1000))
        self.assertAlmostEqualOnAllArrayTypes(sample_entropy, ts, 0.0010314596066622707)

    def test_autocorrelation(self):
        self.assertAlmostEqualOnAllArrayTypes(autocorrelation, [1, 2, 1, 2, 1, 2], -1, 1)
        self.assertAlmostEqualOnAllArrayTypes(autocorrelation, [1, 2, 1, 2, 1, 2], 1, 2)
        self.assertAlmostEqualOnAllArrayTypes(autocorrelation, [1, 2, 1, 2, 1, 2], -1, 3)
        self.assertAlmostEqualOnAllArrayTypes(autocorrelation, [1, 2, 1, 2, 1, 2], 1, 4)
        self.assertAlmostEqualOnAllArrayTypes(autocorrelation, pd.Series([0, 1, 2, 0, 1, 2]), -0.75, 2)
        # Autocorrelation lag is larger than length of the time series
        self.assertIsNanOnAllArrayTypes(autocorrelation, [1, 2, 1, 2, 1, 2], 200)
        self.assertIsNanOnAllArrayTypes(autocorrelation, [np.nan], 0)
        self.assertIsNanOnAllArrayTypes(autocorrelation, [], 0)
        # time series with length 1 has no variance, therefore no result for autocorrelation at lag 0
        self.assertIsNanOnAllArrayTypes(autocorrelation, [1], 0)

    def test_quantile(self):
        self.assertAlmostEqualOnAllArrayTypes(quantile, [1, 1, 1, 3, 4, 7, 9, 11, 13, 13], 1.0, 0.2)
        self.assertAlmostEqualOnAllArrayTypes(quantile, [1, 1, 1, 3, 4, 7, 9, 11, 13, 13], 13, 0.9)
        self.assertAlmostEqualOnAllArrayTypes(quantile, [1, 1, 1, 3, 4, 7, 9, 11, 13, 13], 13, 1.0)
        self.assertAlmostEqualOnAllArrayTypes(quantile, [1], 1, 0.5)
        self.assertIsNanOnAllArrayTypes(quantile, [], 0.5)

    def test_mean_abs_change_quantiles(self):

        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, list(range(10)), 1,
                                              ql=0.1, qh=0.9, isabs=True, f_agg="mean")
        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, list(range(10)), 0,
                                              ql=0.15, qh=0.18, isabs=True, f_agg="mean")
        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, [0, 1, 0, 0, 0], 0.5,
                                              ql=0, qh=1, isabs=True, f_agg="mean")
        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, [0, 1, 0, 0, 0], 0.5,
                                              ql=0.1, qh=1, isabs=True, f_agg="mean")
        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, [0, 1, 0, 0, 0], 0,
                                              ql=0.1, qh=0.6, isabs=True, f_agg="mean")
        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, [0, 1, -9, 0, 0], 5,
                                              ql=0, qh=1, isabs=True, f_agg="mean")
        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, [0, 1, -9, 0, 0], 0.5,
                                              ql=0.1, qh=1, isabs=True, f_agg="mean")
        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, [0, 1, -9, 0, 0, 1, 0], 0.75,
                                              ql=0.1, qh=1, isabs=True, f_agg="mean")

        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, list(range(10)), 1,
                                              ql=0.1, qh=0.9, isabs=False, f_agg="mean")
        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, list(range(10)), 0,
                                              ql=0.15, qh=0.18, isabs=False, f_agg="mean")
        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, [0, 1, 0, 0, 0], 0,
                                              ql=0, qh=1, isabs=False, f_agg="mean")
        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, [0, 1, 0, 0, 0], 0,
                                              ql=0.1, qh=1, isabs=False, f_agg="mean")
        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, [0, 1, 0, 0, 0], 0,
                                              ql=0.1, qh=0.6, isabs=False, f_agg="mean")
        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, [0, 1, -9, 0, 0], 0,
                                              ql=0, qh=1, isabs=False, f_agg="mean")
        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, [0, 1, -9, 0, 0], 0.5,
                                              ql=0.1, qh=1, isabs=False, f_agg="mean")
        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, [0, 1, -9, 0, 0, 1, 0], 0.25,
                                              ql=0.1, qh=1, isabs=False, f_agg="mean")

        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, list(range(10)), 0,
                                              ql=0.1, qh=0.9, isabs=True, f_agg="std")
        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, [0, 1, 0, 0, 0], 0.5,
                                              ql=0, qh=1, isabs=True, f_agg="std")

        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, list(range(10)), 0,
                                              ql=0.1, qh=0.9, isabs=False, f_agg="std")
        self.assertAlmostEqualOnAllArrayTypes(change_quantiles, [0, 1, 0, 1, 0], 1,
                                              ql=0, qh=1, isabs=False, f_agg="std")

    def test_value_count(self):
        self.assertEqualPandasSeriesWrapper(value_count, [1] * 10, 10, value=1)
        self.assertEqualPandasSeriesWrapper(value_count, list(range(10)), 1, value=0)
        self.assertEqualPandasSeriesWrapper(value_count, [1] * 10, 0, value=0)
        self.assertEqualPandasSeriesWrapper(value_count, [np.NaN, 0, 1] * 3, 3, value=0)
        self.assertEqualPandasSeriesWrapper(value_count, [np.NINF, 0, 1] * 3, 3, value=0)
        self.assertEqualPandasSeriesWrapper(value_count, [np.PINF, 0, 1] * 3, 3, value=0)
        self.assertEqualPandasSeriesWrapper(value_count, [0.1, 0.2, 0.3] * 3, 3, value=0.2)
        self.assertEqualPandasSeriesWrapper(value_count, [np.NaN, 0, 1] * 3, 3, value=np.NaN)
        self.assertEqualPandasSeriesWrapper(value_count, [np.NINF, 0, 1] * 3, 3, value=np.NINF)
        self.assertEqualPandasSeriesWrapper(value_count, [np.PINF, 0, 1] * 3, 3, value=np.PINF)

    def test_range_count(self):
        self.assertEqualPandasSeriesWrapper(range_count, [1] * 10, 0, min=1, max=1)
        self.assertEqualPandasSeriesWrapper(range_count, [1] * 10, 0, min=0.9, max=1)
        self.assertEqualPandasSeriesWrapper(range_count, [1] * 10, 10, min=1, max=1.1)
        self.assertEqualPandasSeriesWrapper(range_count, list(range(10)), 9, min=0, max=9)
        self.assertEqualPandasSeriesWrapper(range_count, list(range(10)), 10, min=0, max=10)
        self.assertEqualPandasSeriesWrapper(range_count, list(range(0, -10, -1)), 9, min=-10, max=0)
        self.assertEqualPandasSeriesWrapper(range_count, [np.NaN, np.PINF, np.NINF] + list(range(10)), 10, min=0,
                                            max=10)

    def test_approximate_entropy(self):
        self.assertEqualOnAllArrayTypes(approximate_entropy, [1], 0, m=2, r=0.5)
        self.assertEqualOnAllArrayTypes(approximate_entropy, [1, 2], 0, m=2, r=0.5)
        self.assertEqualOnAllArrayTypes(approximate_entropy, [1, 2, 3], 0, m=2, r=0.5)
        self.assertEqualOnAllArrayTypes(approximate_entropy, [1, 2, 3], 0, m=2, r=0.5)
        self.assertAlmostEqualOnAllArrayTypes(approximate_entropy, [12, 13, 15, 16, 17] * 10, 0.282456191, m=2, r=0.9)
        self.assertRaises(ValueError, approximate_entropy, x=[12, 13, 15, 16, 17] * 10, m=2, r=-0.5)

    def test_max_langevin_fixed_point(self):
        """
        Estimating the intrinsic velocity of a dissipative soliton
        """
        default_params = {"m": 3, "r": 30}

        # active Brownian motion
        ds = velocity(tau=3.8, delta_t=0.05, R=3e-4, seed=0)
        v = ds.simulate(100000, v0=np.zeros(1))
        v0 = max_langevin_fixed_point(v[:, 0], **default_params)
        self.assertLess(abs(ds.deterministic - v0), 0.001)

        # Brownian motion
        ds = velocity(tau=2.0 / 0.3 - 3.8, delta_t=0.05, R=3e-4, seed=0)
        v = ds.simulate(10000, v0=np.zeros(1))
        v0 = max_langevin_fixed_point(v[:, 0], **default_params)
        self.assertLess(v0, 0.001)

    def test_linear_trend(self):
        # check linear up trend
        x = range(10)
        param = [{"attr": "pvalue"}, {"attr": "rvalue"}, {"attr": "intercept"}, {"attr": "slope"}, {"attr": "stderr"}]
        res = linear_trend(x, param)

        res = pd.Series(dict(res))

        expected_index = ["attr_\"pvalue\"", "attr_\"intercept\"",
                          "attr_\"rvalue\"", "attr_\"slope\"",
                          "attr_\"stderr\""]

        self.assertEqual(len(res), 5)
        self.assertCountEqual(list(res.index), expected_index)
        self.assertAlmostEqual(res["attr_\"pvalue\""], 0)
        self.assertAlmostEqual(res["attr_\"stderr\""], 0)
        self.assertAlmostEqual(res["attr_\"intercept\""], 0)
        self.assertAlmostEqual(res["attr_\"slope\""], 1.0)

        # check p value for random trend
        np.random.seed(42)
        x = np.random.uniform(size=100)
        param = [{"attr": "rvalue"}]
        res = linear_trend(x, param)

        res = pd.Series(dict(res))

        self.assertLess(abs(res["attr_\"rvalue\""]), 0.1)

        # check slope and intercept decreasing trend with intercept
        x = [42 - 2 * x for x in range(10)]
        param = [{"attr": "intercept"}, {"attr": "slope"}]
        res = linear_trend(x, param)

        res = pd.Series(dict(res))

        self.assertAlmostEqual(res["attr_\"intercept\""], 42)
        self.assertAlmostEqual(res["attr_\"slope\""], -2)

    def test__aggregate_on_chunks(self):
        self.assertListEqual(_aggregate_on_chunks(x=pd.Series([0, 1, 2, 3]), f_agg="max", chunk_len=2), [1, 3])
        self.assertListEqual(_aggregate_on_chunks(x=pd.Series([1, 1, 3, 3]), f_agg="max", chunk_len=2), [1, 3])

        self.assertListEqual(_aggregate_on_chunks(x=pd.Series([0, 1, 2, 3]), f_agg="min", chunk_len=2), [0, 2])
        self.assertListEqual(_aggregate_on_chunks(x=pd.Series([0, 1, 2, 3, 5]), f_agg="min", chunk_len=2), [0, 2, 5])

        self.assertListEqual(_aggregate_on_chunks(x=pd.Series([0, 1, 2, 3]), f_agg="mean", chunk_len=2),
                             [0.5, 2.5])
        self.assertListEqual(_aggregate_on_chunks(x=pd.Series([0, 1, 0, 4, 5]), f_agg="mean", chunk_len=2),
                             [0.5, 2, 5])
        self.assertListEqual(_aggregate_on_chunks(x=pd.Series([0, 1, 0, 4, 5]), f_agg="mean", chunk_len=3),
                             [1 / 3, 4.5])

        self.assertListEqual(_aggregate_on_chunks(x=pd.Series([0, 1, 2, 3, 5, -2]),
                                                  f_agg="median", chunk_len=2), [0.5, 2.5, 1.5])
        self.assertListEqual(_aggregate_on_chunks(x=pd.Series([-10, 5, 3, -3, 4, -6]),
                                                  f_agg="median", chunk_len=3), [3, -3])
        self.assertListEqual(_aggregate_on_chunks(x=pd.Series([0, 1, 2, np.NaN, 5]),
                                                  f_agg="median", chunk_len=2), [0.5, 2, 5])

    def test_agg_linear_trend(self):
        x = pd.Series(range(9), index=range(9))
        param = [{"attr": "intercept", "chunk_len": 3, "f_agg": "max"},
                 {"attr": "slope", "chunk_len": 3, "f_agg": "max"},
                 {"attr": "intercept", "chunk_len": 3, "f_agg": "min"},
                 {"attr": "slope", "chunk_len": 3, "f_agg": "min"},
                 {"attr": "intercept", "chunk_len": 3, "f_agg": "mean"},
                 {"attr": "slope", "chunk_len": 3, "f_agg": "mean"},
                 {"attr": "intercept", "chunk_len": 3, "f_agg": "median"},
                 {"attr": "slope", "chunk_len": 3, "f_agg": "median"}]
        expected_index = ['attr_"intercept"__chunk_len_3__f_agg_"max"',
                          'attr_"slope"__chunk_len_3__f_agg_"max"',
                          'attr_"intercept"__chunk_len_3__f_agg_"min"',
                          'attr_"slope"__chunk_len_3__f_agg_"min"',
                          'attr_"intercept"__chunk_len_3__f_agg_"mean"',
                          'attr_"slope"__chunk_len_3__f_agg_"mean"',
                          'attr_"intercept"__chunk_len_3__f_agg_"median"',
                          'attr_"slope"__chunk_len_3__f_agg_"median"']

        res = agg_linear_trend(x=x, param=param)

        res = pd.Series(dict(res))
        self.assertEqual(len(res), 8)
        self.maxDiff = 2000
        self.assertCountEqual(list(res.index), expected_index)
        self.assertAlmostEqual(res['attr_"intercept"__chunk_len_3__f_agg_"max"'], 2)
        self.assertAlmostEqual(res['attr_"slope"__chunk_len_3__f_agg_"max"'], 3)
        self.assertAlmostEqual(res['attr_"intercept"__chunk_len_3__f_agg_"min"'], 0)
        self.assertAlmostEqual(res['attr_"slope"__chunk_len_3__f_agg_"min"'], 3)
        self.assertAlmostEqual(res['attr_"intercept"__chunk_len_3__f_agg_"mean"'], 1)
        self.assertAlmostEqual(res['attr_"slope"__chunk_len_3__f_agg_"mean"'], 3)
        self.assertAlmostEqual(res['attr_"intercept"__chunk_len_3__f_agg_"median"'], 1)
        self.assertAlmostEqual(res['attr_"slope"__chunk_len_3__f_agg_"median"'], 3)

        x = pd.Series([np.NaN, np.NaN, np.NaN, -3, -3, -3])
        res = agg_linear_trend(x=x, param=param)

        res = pd.Series(dict(res))

        self.assertIsNaN(res['attr_"intercept"__chunk_len_3__f_agg_"max"'])
        self.assertIsNaN(res['attr_"slope"__chunk_len_3__f_agg_"max"'])
        self.assertIsNaN(res['attr_"intercept"__chunk_len_3__f_agg_"min"'])
        self.assertIsNaN(res['attr_"slope"__chunk_len_3__f_agg_"min"'])
        self.assertIsNaN(res['attr_"intercept"__chunk_len_3__f_agg_"mean"'])
        self.assertIsNaN(res['attr_"slope"__chunk_len_3__f_agg_"mean"'])
        self.assertIsNaN(res['attr_"intercept"__chunk_len_3__f_agg_"median"'])
        self.assertIsNaN(res['attr_"slope"__chunk_len_3__f_agg_"median"'])

        x = pd.Series([np.NaN, np.NaN, -3, -3, -3, -3])
        res = agg_linear_trend(x=x, param=param)

        res = pd.Series(dict(res))

        self.assertAlmostEqual(res['attr_"intercept"__chunk_len_3__f_agg_"max"'], -3)
        self.assertAlmostEqual(res['attr_"slope"__chunk_len_3__f_agg_"max"'], 0)
        self.assertAlmostEqual(res['attr_"intercept"__chunk_len_3__f_agg_"min"'], -3)
        self.assertAlmostEqual(res['attr_"slope"__chunk_len_3__f_agg_"min"'], 0)
        self.assertAlmostEqual(res['attr_"intercept"__chunk_len_3__f_agg_"mean"'], -3)
        self.assertAlmostEqual(res['attr_"slope"__chunk_len_3__f_agg_"mean"'], 0)
        self.assertAlmostEqual(res['attr_"intercept"__chunk_len_3__f_agg_"median"'], -3)
        self.assertAlmostEqual(res['attr_"slope"__chunk_len_3__f_agg_"median"'], 0)

    def test_energy_ratio_by_chunks(self):
        x = pd.Series(range(90), index=range(90))
        param = [{"num_segments": 6, "segment_focus": i} for i in range(6)]
        output = energy_ratio_by_chunks(x=x, param=param)

        self.assertAlmostEqual(output[0][1], 0.0043, places=3)
        self.assertAlmostEqual(output[1][1], 0.0316, places=3)
        self.assertAlmostEqual(output[2][1], 0.0871, places=3)
        self.assertAlmostEqual(output[3][1], 0.1709, places=3)
        self.assertAlmostEqual(output[4][1], 0.2829, places=3)
        self.assertAlmostEqual(output[5][1], 0.4232, places=3)

        # Sum of the ratios should be 1.0
        sum = 0.0
        for name, dat in output:
            sum = sum + dat
        self.assertAlmostEqual(sum, 1.0)

        x = pd.Series(1, index=range(10))
        param = [{"num_segments": 3, "segment_focus": i} for i in range(3)]
        output = energy_ratio_by_chunks(x=x, param=param)

        self.assertAlmostEqual(output[0][1], 0.4, places=3)
        self.assertAlmostEqual(output[1][1], 0.3, places=3)
        self.assertAlmostEqual(output[2][1], 0.3, places=3)

        # Sum of the ratios should be 1.0
        sum = 0.0
        for name, dat in output:
            sum = sum + dat
        self.assertAlmostEqual(sum, 1.0)

        x = pd.Series(0, index=range(10))
        param = [{"num_segments": 3, "segment_focus": i} for i in range(3)]
        output = energy_ratio_by_chunks(x=x, param=param)
        self.assertIsNaN(output[0][1])
        self.assertIsNaN(output[1][1])
        self.assertIsNaN(output[2][1])

    def test_linear_trend_timewise_hours(self):
        """Test linear_trend_timewise function with hour intervals."""
        x = pd.Series(
            [0, 1, 3, 6],
            index=pd.DatetimeIndex([
                '2018-01-01 04:00:00', '2018-01-01 05:00:00',
                '2018-01-01 07:00:00', '2018-01-01 10:00:00'
            ]),
        )

        param = [{"attr": "pvalue"}, {"attr": "rvalue"}, {"attr": "intercept"}, {"attr": "slope"}, {"attr": "stderr"}]
        res = linear_trend_timewise(x, param)

        res = pd.Series(dict(res))

        expected_index = ["attr_\"pvalue\"", "attr_\"intercept\"",
                          "attr_\"rvalue\"", "attr_\"slope\"",
                          "attr_\"stderr\""]

        self.assertEqual(len(res), 5)
        self.assertCountEqual(list(res.index), expected_index)
        self.assertAlmostEqual(res["attr_\"pvalue\""], 0, places=3)
        self.assertAlmostEqual(res["attr_\"stderr\""], 0, places=3)
        self.assertAlmostEqual(res["attr_\"intercept\""], 0, places=3)
        self.assertAlmostEqual(res["attr_\"slope\""], 1.0, places=3)

    def test_linear_trend_timewise_days(self):
        """Test linear_trend_timewise function with day intervals."""
        # Try with different days
        x = pd.Series(
            [0, 24, 48, 72],
            index=pd.DatetimeIndex([
                '2018-01-01 04:00:00', '2018-01-02 04:00:00',
                '2018-01-03 04:00:00', '2018-01-04 04:00:00'
            ]),
        )

        param = [{"attr": "pvalue"}, {"attr": "rvalue"}, {"attr": "intercept"}, {"attr": "slope"}, {"attr": "stderr"}]
        res = linear_trend_timewise(x, param)

        res = pd.Series(dict(res))

        self.assertAlmostEqual(res["attr_\"pvalue\""], 0, places=3)
        self.assertAlmostEqual(res["attr_\"stderr\""], 0, places=3)
        self.assertAlmostEqual(res["attr_\"intercept\""], 0, places=3)
        self.assertAlmostEqual(res["attr_\"slope\""], 1.0, places=3)

    def test_linear_trend_timewise_seconds(self):
        """Test linear_trend_timewise function with second intervals."""
        # Try with different days
        x = pd.Series(
            [0, 1 / float(3600), 2 / float(3600), 3 / float(3600)],
            index=pd.DatetimeIndex([
                '2018-01-01 04:00:01', '2018-01-01 04:00:02',
                '2018-01-01 04:00:03', '2018-01-01 04:00:04'
            ]),
        )

        param = [{"attr": "pvalue"}, {"attr": "rvalue"}, {"attr": "intercept"}, {"attr": "slope"}, {"attr": "stderr"}]
        res = linear_trend_timewise(x, param)

        res = pd.Series(dict(res))

        self.assertAlmostEqual(res["attr_\"pvalue\""], 0, places=3)
        self.assertAlmostEqual(res["attr_\"stderr\""], 0, places=3)
        self.assertAlmostEqual(res["attr_\"intercept\""], 0, places=3)
        self.assertAlmostEqual(res["attr_\"slope\""], 1.0, places=3)

    def test_linear_trend_timewise_years(self):
        """Test linear_trend_timewise function with year intervals."""
        # Try with different days
        x = pd.Series(
            [0, 365 * 24, 365 * 48, 365 * 72 + 24],  # Add 24 to the last one since it's a leap year
            index=pd.DatetimeIndex([
                '2018-01-01 04:00:00', '2019-01-01 04:00:00',
                '2020-01-01 04:00:00', '2021-01-01 04:00:00'
            ]),
        )

        param = [{"attr": "pvalue"}, {"attr": "rvalue"}, {"attr": "intercept"}, {"attr": "slope"}, {"attr": "stderr"}]
        res = linear_trend_timewise(x, param)

        res = pd.Series(dict(res))

        self.assertAlmostEqual(res["attr_\"pvalue\""], 0, places=3)
        self.assertAlmostEqual(res["attr_\"stderr\""], 0, places=3)
        self.assertAlmostEqual(res["attr_\"intercept\""], 0, places=3)
        self.assertAlmostEqual(res["attr_\"slope\""], 1.0, places=3)

    def test_change_quantiles(self):
        """Test change_quantiles function when changing from `sum` to `np.sum`."""
        np.random.seed(0)
        res = change_quantiles(np.random.rand(10000) * 1000, 0.1, 0.2, False, 'mean')
        self.assertAlmostEqual(res, -0.9443846621365727)

    def test_count_above(self):
        self.assertEqualPandasSeriesWrapper(count_above, [1] * 10, 1, t=1)
        self.assertEqualPandasSeriesWrapper(count_above, list(range(10)), 1, t=0)
        self.assertEqualPandasSeriesWrapper(count_above, list(range(10)), 0.5, t=5)
        self.assertEqualPandasSeriesWrapper(count_above, [0.1, 0.2, 0.3] * 3, 2 / 3, t=0.2)
        self.assertEqualPandasSeriesWrapper(count_above, [np.NaN, 0, 1] * 3, 2 / 3, t=0)
        self.assertEqualPandasSeriesWrapper(count_above, [np.NINF, 0, 1] * 3, 2 / 3, t=0)
        self.assertEqualPandasSeriesWrapper(count_above, [np.PINF, 0, 1] * 3, 1, t=0)
        self.assertEqualPandasSeriesWrapper(count_above, [np.NaN, 0, 1] * 3, 0, t=np.NaN)
        self.assertEqualPandasSeriesWrapper(count_above, [np.NINF, 0, np.PINF] * 3, 1, t=np.NINF)
        self.assertEqualPandasSeriesWrapper(count_above, [np.PINF, 0, 1] * 3, 1 / 3, t=np.PINF)

    def test_count_below(self):
        self.assertEqualPandasSeriesWrapper(count_below, [1] * 10, 1, t=1)
        self.assertEqualPandasSeriesWrapper(count_below, list(range(10)), 1 / 10, t=0)
        self.assertEqualPandasSeriesWrapper(count_below, list(range(10)), 6 / 10, t=5)
        self.assertEqualPandasSeriesWrapper(count_below, [0.1, 0.2, 0.3] * 3, 2 / 3, t=0.2)
        self.assertEqualPandasSeriesWrapper(count_below, [np.NaN, 0, 1] * 3, 1 / 3, t=0)
        self.assertEqualPandasSeriesWrapper(count_below, [np.NINF, 0, 1] * 3, 2 / 3, t=0)
        self.assertEqualPandasSeriesWrapper(count_below, [np.PINF, 0, 1] * 3, 1 / 3, t=0)
        self.assertEqualPandasSeriesWrapper(count_below, [np.NaN, 0, 1] * 3, 0, t=np.NaN)
        self.assertEqualPandasSeriesWrapper(count_below, [np.NINF, 0, np.PINF] * 3, 1 / 3, t=np.NINF)
        self.assertEqualPandasSeriesWrapper(count_below, [np.PINF, 0, 1] * 3, 1, t=np.PINF)

    def test_benford_correlation(self):
        # A test with list of random values
        np.random.seed(42)
        random_list = np.random.uniform(size=100)

        # Fibonacci series is known to match the Newcomb-Benford's Distribution
        fibonacci_list = [0, 1]
        for i in range(2, 200):
            fibonacci_list.append(fibonacci_list[i - 1] + fibonacci_list[i - 2])

        # A list of equally distributed digits (returns NaN)
        equal_list = [1, 2, 3, 4, 5, 6, 7, 8, 9]

        # A list containing NaN
        list_with_nan = [1.354, 0.058, 0.055, 0.99, 3.15, np.nan, 0.3, 2.3, 0, 0.59, 0.74]

        self.assertAlmostEqual(benford_correlation(random_list), 0.39458056)
        self.assertAlmostEqual(benford_correlation(fibonacci_list), 0.998003988)
        self.assertAlmostEqual(benford_correlation(list_with_nan), 0.10357511)
        self.assertIsNaN(benford_correlation(equal_list))


class FriedrichTestCase(TestCase):

    def test_estimate_friedrich_coefficients(self):
        """
        Estimate friedrich coefficients
        """
        default_params = {"m": 3, "r": 30}

        # active Brownian motion
        ds = velocity(tau=3.8, delta_t=0.05, R=3e-4, seed=0)
        v = ds.simulate(10000, v0=np.zeros(1))
        coeff = _estimate_friedrich_coefficients(v[:, 0], **default_params)
        self.assertLess(abs(coeff[-1]), 0.0001)

        # Brownian motion
        ds = velocity(tau=2.0 / 0.3 - 3.8, delta_t=0.05, R=3e-4, seed=0)
        v = ds.simulate(10000, v0=np.zeros(1))
        coeff = _estimate_friedrich_coefficients(v[:, 0], **default_params)
        self.assertLess(abs(coeff[-1]), 0.0001)

    def test_friedrich_coefficients(self):
        # Test binning error returns vector of NaNs
        param = [{"coeff": coeff, "m": 2, "r": 30} for coeff in range(4)]
        x = np.zeros(100)
        res = pd.Series(dict(friedrich_coefficients(x, param)))

        expected_index = ["coeff_0__m_2__r_30", "coeff_1__m_2__r_30", "coeff_2__m_2__r_30", "coeff_3__m_2__r_30"]
        self.assertCountEqual(list(res.index), expected_index)
        self.assertTrue(np.sum(np.isnan(res)), 3)

    def test_friedrich_number_of_returned_features_is_equal_to_number_of_parameters(self):
        """ unit test for issue 501 """
        param = [{'m': 3, 'r': 5, 'coeff': 2}, {'m': 3, 'r': 5, 'coeff': 3}, {'m': 3, 'r': 2, 'coeff': 3}]
        x = np.zeros(100)
        res = pd.Series(dict(friedrich_coefficients(x, param)))

        expected_index = ["coeff_2__m_3__r_5", "coeff_3__m_3__r_5", "coeff_3__m_3__r_2"]
        self.assertCountEqual(list(res.index), expected_index)
        self.assertTrue(np.sum(np.isnan(res)), 3)

    def test_friedrich_equal_to_snapshot(self):
        param = [{"coeff": coeff, "m": 2, "r": 30} for coeff in range(4)]
        x = np.array([-0.53, -0.61, -1.26, -0.88, -0.34, 0.58, 2.86, -0.47, 0.78,
                      -0.45, -0.27, 0.43, 1.72, 0.26, 1.02, -0.09, 0.65, 1.49,
                      -0.95, -1.02, -0.64, -1.63, -0.71, -0.43, -1.69, 0.05, 1.58,
                      1.1, 0.55, -1.02])

        res = pd.Series(dict(friedrich_coefficients(x, param)))

        self.assertAlmostEqual(res['coeff_0__m_2__r_30'], -0.24536975738843042)
        self.assertAlmostEqual(res['coeff_1__m_2__r_30'], -0.533309548662685)
        self.assertAlmostEqual(res['coeff_2__m_2__r_30'], 0.2759399238199404)



# -*- coding: utf-8 -*-
# This file as well as the whole tsfresh package are licenced under the MIT licence (see the LICENCE.txt)
# Maximilian Christ (maximilianchrist.com), Blue Yonder Gmbh, 2016

import pandas as pd
from tsfresh.feature_extraction import extract_features
from tsfresh import defaults
from tsfresh.feature_selection import select_features
from tsfresh.utilities.dataframe_functions import restrict_input_to_index, impute, get_ids


def extract_relevant_features(timeseries_container, y, X=None,
                              default_fc_parameters=None,
                              kind_to_fc_parameters=None,
                              column_id=None, column_sort=None, column_kind=None, column_value=None,
                              show_warnings=defaults.SHOW_WARNINGS,
                              disable_progressbar=defaults.DISABLE_PROGRESSBAR,
                              profile=defaults.PROFILING,
                              profiling_filename=defaults.PROFILING_FILENAME,
                              profiling_sorting=defaults.PROFILING_SORTING,
                              test_for_binary_target_binary_feature=defaults.TEST_FOR_BINARY_TARGET_BINARY_FEATURE,
                              test_for_binary_target_real_feature=defaults.TEST_FOR_BINARY_TARGET_REAL_FEATURE,
                              test_for_real_target_binary_feature=defaults.TEST_FOR_REAL_TARGET_BINARY_FEATURE,
                              test_for_real_target_real_feature=defaults.TEST_FOR_REAL_TARGET_REAL_FEATURE,
                              fdr_level=defaults.FDR_LEVEL,
                              hypotheses_independent=defaults.HYPOTHESES_INDEPENDENT,
                              n_jobs=defaults.N_PROCESSES,
                              distributor=None,
                              chunksize=defaults.CHUNKSIZE,
                              ml_task='auto'):
    """
    High level convenience function to extract time series features from `timeseries_container`. Then return feature
    matrix `X` possibly augmented with relevant features with respect to target vector `y`.

    For more details see the documentation of :func:`~tsfresh.feature_extraction.extraction.extract_features` and
    :func:`~tsfresh.feature_selection.selection.select_features`.

    Examples
    ========

    >>> from tsfresh.examples import load_robot_execution_failures
    >>> from tsfresh import extract_relevant_features
    >>> df, y = load_robot_execution_failures()
    >>> X = extract_relevant_features(df, y, column_id='id', column_sort='time')

    :param timeseries_container: The pandas.DataFrame with the time series to compute the features for, or a
            dictionary of pandas.DataFrames.
            See :func:`~tsfresh.feature_extraction.extraction.extract_features`.

    :param X: A DataFrame containing additional features
    :type X: pandas.DataFrame

    :param y: The target vector
    :type y: pandas.Series

    :param default_fc_parameters: mapping from feature calculator names to parameters. Only those names
           which are keys in this dict will be calculated. See the class:`ComprehensiveFCParameters` for
           more information.
    :type default_fc_parameters: dict

    :param kind_to_fc_parameters: mapping from kind names to objects of the same type as the ones for
            default_fc_parameters. If you put a kind as a key here, the fc_parameters
            object (which is the value), will be used instead of the default_fc_parameters.
    :type kind_to_fc_parameters: dict

    :param column_id: The name of the id column to group by. Please see :ref:`data-formats-label`.
    :type column_id: str

    :param column_sort: The name of the sort column. Please see :ref:`data-formats-label`.
    :type column_sort: str

    :param column_kind: The name of the column keeping record on the kind of the value.
            Please see :ref:`data-formats-label`.
    :type column_kind: str

    :param column_value: The name for the column keeping the value itself. Please see :ref:`data-formats-label`.
    :type column_value: str

    :param chunksize: The size of one chunk that is submitted to the worker
        process for the parallelisation.  Where one chunk is defined as a
        singular time series for one id and one kind. If you set the chunksize
        to 10, then it means that one task is to calculate all features for 10
        time series.  If it is set it to None, depending on distributor,
        heuristics are used to find the optimal chunksize. If you get out of
        memory exceptions, you can try it with the dask distributor and a
        smaller chunksize.
    :type chunksize: None or int

    :param n_jobs: The number of processes to use for parallelization. If zero, no parallelization is used.
    :type n_jobs: int

    :param distributor: Advanced parameter: set this to a class name that you want to use as a
             distributor. See the utilities/distribution.py for more information. Leave to None, if you want
             TSFresh to choose the best distributor.
    :type distributor: class

    :param show_warnings: Show warnings during the feature extraction (needed for debugging of calculators).
    :type show_warnings: bool

    :param disable_progressbar: Do not show a progressbar while doing the calculation.
    :type disable_progressbar: bool

    :param profile: Turn on profiling during feature extraction
    :type profile: bool

    :param profiling_sorting: How to sort the profiling results (see the documentation of the profiling package for
           more information)
    :type profiling_sorting: basestring

    :param profiling_filename: Where to save the profiling results.
    :type profiling_filename: basestring

    :param test_for_binary_target_binary_feature: Which test to be used for binary target, binary feature
                                                  (currently unused)
    :type test_for_binary_target_binary_feature: str

    :param test_for_binary_target_real_feature: Which test to be used for binary target, real feature
    :type test_for_binary_target_real_feature: str

    :param test_for_real_target_binary_feature: Which test to be used for real target, binary feature (currently unused)
    :type test_for_real_target_binary_feature: str

    :param test_for_real_target_real_feature: Which test to be used for real target, real feature (currently unused)
    :type test_for_real_target_real_feature: str

    :param fdr_level: The FDR level that should be respected, this is the theoretical expected percentage of irrelevant
                      features among all created features.
    :type fdr_level: float

    :param hypotheses_independent: Can the significance of the features be assumed to be independent?
                                   Normally, this should be set to False as the features are never
                                   independent (e.g. mean and median)
    :type hypotheses_independent: bool

    :param ml_task: The intended machine learning task. Either `'classification'`, `'regression'` or `'auto'`.
                    Defaults to `'auto'`, meaning the intended task is inferred from `y`.
                    If `y` has a boolean, integer or object dtype, the task is assumend to be classification,
                    else regression.
    :type ml_task: str

    :return: Feature matrix X, possibly extended with relevant time series features.
    """

    assert isinstance(y, pd.Series), "y needs to be a pandas.Series, received type: {}.".format(type(y))
    assert len(set(y)) > 1, "Feature selection is only possible if more than 1 label/class is provided"

    if X is not None:
        timeseries_container = restrict_input_to_index(timeseries_container, column_id, X.index)

    ids_container = get_ids(df_or_dict=timeseries_container, column_id=column_id)
    ids_y = set(y.index)
    if ids_container != ids_y:
        if len(ids_container - ids_y) > 0:
            raise ValueError("The following ids are in the time series container but are missing in y: "
                             "{}".format(ids_container - ids_y))
        if len(ids_y - ids_container) > 0:
            raise ValueError("The following ids are in y but are missing inside the time series container: "
                             "{}".format(ids_y - ids_container))

    X_ext = extract_features(timeseries_container,
                             default_fc_parameters=default_fc_parameters,
                             kind_to_fc_parameters=kind_to_fc_parameters,
                             show_warnings=show_warnings,
                             disable_progressbar=disable_progressbar,
                             profile=profile,
                             profiling_filename=profiling_filename,
                             profiling_sorting=profiling_sorting,
                             n_jobs=n_jobs,
                             column_id=column_id, column_sort=column_sort,
                             column_kind=column_kind, column_value=column_value,
                             distributor=distributor,
                             impute_function=impute)

    X_sel = select_features(X_ext, y,
                            test_for_binary_target_binary_feature=test_for_binary_target_binary_feature,
                            test_for_binary_target_real_feature=test_for_binary_target_real_feature,
                            test_for_real_target_binary_feature=test_for_real_target_binary_feature,
                            test_for_real_target_real_feature=test_for_real_target_real_feature,
                            fdr_level=fdr_level, hypotheses_independent=hypotheses_independent,
                            n_jobs=n_jobs,
                            show_warnings=show_warnings,
                            chunksize=chunksize,
                            ml_task=ml_task)

    if X is None:
        X = X_sel
    else:
        X = pd.merge(X, X_sel, left_index=True, right_index=True, how="left")

    return X



# -*- coding: utf-8 -*-
# This file as well as the whole tsfresh package are licenced under the MIT licence (see the LICENCE.txt)
# Maximilian Christ (maximilianchrist.com), Blue Yonder Gmbh, 2016
"""
This module contains the filtering process for the extracted features. The filtering procedure can also be used on
other features that are not based on time series.
"""

import pandas as pd
import numpy as np
from tsfresh import defaults
from tsfresh.utilities.dataframe_functions import check_for_nans_in_columns
from tsfresh.feature_selection.relevance import calculate_relevance_table


def select_features(X, y, test_for_binary_target_binary_feature=defaults.TEST_FOR_BINARY_TARGET_BINARY_FEATURE,
                    test_for_binary_target_real_feature=defaults.TEST_FOR_BINARY_TARGET_REAL_FEATURE,
                    test_for_real_target_binary_feature=defaults.TEST_FOR_REAL_TARGET_BINARY_FEATURE,
                    test_for_real_target_real_feature=defaults.TEST_FOR_REAL_TARGET_REAL_FEATURE,
                    fdr_level=defaults.FDR_LEVEL, hypotheses_independent=defaults.HYPOTHESES_INDEPENDENT,
                    n_jobs=defaults.N_PROCESSES, show_warnings=defaults.SHOW_WARNINGS,
                    chunksize=defaults.CHUNKSIZE, ml_task='auto'):
    """
    Check the significance of all features (columns) of feature matrix X and return a possibly reduced feature matrix
    only containing relevant features.

    The feature matrix must be a pandas.DataFrame in the format:

        +-------+-----------+-----------+-----+-----------+
        | index | feature_1 | feature_2 | ... | feature_N |
        +=======+===========+===========+=====+===========+
        | A     | ...       | ...       | ... | ...       |
        +-------+-----------+-----------+-----+-----------+
        | B     | ...       | ...       | ... | ...       |
        +-------+-----------+-----------+-----+-----------+
        | ...   | ...       | ...       | ... | ...       |
        +-------+-----------+-----------+-----+-----------+
        | ...   | ...       | ...       | ... | ...       |
        +-------+-----------+-----------+-----+-----------+
        | ...   | ...       | ...       | ... | ...       |
        +-------+-----------+-----------+-----+-----------+


    Each column will be handled as a feature and tested for its significance to the target.

    The target vector must be a pandas.Series or numpy.array in the form

        +-------+--------+
        | index | target |
        +=======+========+
        | A     | ...    |
        +-------+--------+
        | B     | ...    |
        +-------+--------+
        | .     | ...    |
        +-------+--------+
        | .     | ...    |
        +-------+--------+

    and must contain all id's that are in the feature matrix. If y is a numpy.array without index, it is assumed
    that y has the same order and length than X and the rows correspond to each other.

    Examples
    ========

    >>> from tsfresh.examples import load_robot_execution_failures
    >>> from tsfresh import extract_features, select_features
    >>> df, y = load_robot_execution_failures()
    >>> X_extracted = extract_features(df, column_id='id', column_sort='time')
    >>> X_selected = select_features(X_extracted, y)

    :param X: Feature matrix in the format mentioned before which will be reduced to only the relevant features.
              It can contain both binary or real-valued features at the same time.
    :type X: pandas.DataFrame

    :param y: Target vector which is needed to test which features are relevant. Can be binary or real-valued.
    :type y: pandas.Series or numpy.ndarray

    :param test_for_binary_target_binary_feature: Which test to be used for binary target, binary feature
                                                  (currently unused)
    :type test_for_binary_target_binary_feature: str

    :param test_for_binary_target_real_feature: Which test to be used for binary target, real feature
    :type test_for_binary_target_real_feature: str

    :param test_for_real_target_binary_feature: Which test to be used for real target, binary feature (currently unused)
    :type test_for_real_target_binary_feature: str

    :param test_for_real_target_real_feature: Which test to be used for real target, real feature (currently unused)
    :type test_for_real_target_real_feature: str

    :param fdr_level: The FDR level that should be respected, this is the theoretical expected percentage of irrelevant
                      features among all created features.
    :type fdr_level: float

    :param hypotheses_independent: Can the significance of the features be assumed to be independent?
                                   Normally, this should be set to False as the features are never
                                   independent (e.g. mean and median)
    :type hypotheses_independent: bool

    :param n_jobs: Number of processes to use during the p-value calculation
    :type n_jobs: int

    :param show_warnings: Show warnings during the p-value calculation (needed for debugging of calculators).
    :type show_warnings: bool

    :param chunksize: The size of one chunk that is submitted to the worker
        process for the parallelisation.  Where one chunk is defined as
        the data for one feature. If you set the chunksize
        to 10, then it means that one task is to filter 10 features.
        If it is set it to None, depending on distributor,
        heuristics are used to find the optimal chunksize. If you get out of
        memory exceptions, you can try it with the dask distributor and a
        smaller chunksize.
    :type chunksize: None or int

    :param ml_task: The intended machine learning task. Either `'classification'`, `'regression'` or `'auto'`.
                    Defaults to `'auto'`, meaning the intended task is inferred from `y`.
                    If `y` has a boolean, integer or object dtype, the task is assumend to be classification,
                    else regression.
    :type ml_task: str

    :return: The same DataFrame as X, but possibly with reduced number of columns ( = features).
    :rtype: pandas.DataFrame

    :raises: ``ValueError`` when the target vector does not fit to the feature matrix
             or `ml_task` is not one of `'auto'`, `'classification'` or `'regression'`.
    """
    assert isinstance(X, pd.DataFrame), "Please pass features in X as pandas.DataFrame."
    check_for_nans_in_columns(X)
    assert isinstance(y, (pd.Series, np.ndarray)), "The type of target vector y must be one of: " \
                                                   "pandas.Series, numpy.ndarray"
    assert len(y) > 1, "y must contain at least two samples."
    assert len(X) == len(y), "X and y must contain the same number of samples."
    assert len(set(y)) > 1, "Feature selection is only possible if more than 1 label/class is provided"

    if isinstance(y, pd.Series) and set(X.index) != set(y.index):
        raise ValueError("Index of X and y must be identical if provided")

    if isinstance(y, np.ndarray):
        y = pd.Series(y, index=X.index)

    relevance_table = calculate_relevance_table(
        X, y, ml_task=ml_task, n_jobs=n_jobs, show_warnings=show_warnings, chunksize=chunksize,
        test_for_binary_target_real_feature=test_for_binary_target_real_feature,
        fdr_level=fdr_level, hypotheses_independent=hypotheses_independent,
    )

    relevant_features = relevance_table[relevance_table.relevant].feature

    return X.loc[:, relevant_features]



# -*- coding: utf-8 -*-
# This file as well as the whole tsfresh package are licenced under the MIT licence (see the LICENCE.txt)
# Maximilian Christ (maximilianchrist.com), Blue Yonder Gmbh, 2016

import os
import tempfile

import numpy as np
import pandas as pd
from mock import Mock

from tests.fixtures import DataTestCase
from tsfresh.feature_extraction.extraction import extract_features
from tsfresh.feature_extraction.settings import ComprehensiveFCParameters
from tsfresh.utilities.distribution import IterableDistributorBaseClass, MapDistributor


class ExtractionTestCase(DataTestCase):
    """The unit tests in this module make sure if the time series features are created properly"""

    def setUp(self):
        self.n_jobs = 1
        self.directory = tempfile.gettempdir()

    def test_extract_features(self):
        # todo: implement more methods and test more aspects
        df = self.create_test_data_sample()
        extracted_features = extract_features(df, column_id="id", column_sort="sort",
                                              column_kind="kind", column_value="val",
                                              n_jobs=self.n_jobs)
        self.assertIsInstance(extracted_features, pd.DataFrame)
        self.assertTrue(np.all(extracted_features.a__maximum == np.array([71, 77])))
        self.assertTrue(np.all(extracted_features.a__sum_values == np.array([691, 1017])))
        self.assertTrue(np.all(extracted_features.a__abs_energy == np.array([32211, 63167])))
        self.assertTrue(np.all(extracted_features.b__sum_values == np.array([757, 695])))
        self.assertTrue(np.all(extracted_features.b__minimum == np.array([3, 1])))
        self.assertTrue(np.all(extracted_features.b__abs_energy == np.array([36619, 35483])))
        self.assertTrue(np.all(extracted_features.b__mean == np.array([37.85, 34.75])))
        self.assertTrue(np.all(extracted_features.b__median == np.array([39.5, 28.0])))

        df_sts = self.create_one_valued_time_series()
        extracted_features_sts = extract_features(df_sts, column_id="id", column_sort="sort",
                                                  column_kind="kind", column_value="val",
                                                  n_jobs=self.n_jobs)

        self.assertIsInstance(extracted_features_sts, pd.DataFrame)
        self.assertTrue(np.all(extracted_features_sts.a__maximum == np.array([1.0, 6.0])))
        self.assertTrue(np.all(extracted_features_sts.a__sum_values == np.array([1.0, 11.0])))
        self.assertTrue(np.all(extracted_features_sts.a__count_above_mean == np.array([0, 1])))

    def test_extract_features_uses_only_kind_to_fc_settings(self):
        df = self.create_test_data_sample()
        extracted_features = extract_features(df, column_id="id", column_sort="sort", column_kind="kind",
                                              column_value="val", n_jobs=self.n_jobs,
                                              kind_to_fc_parameters={"a": {"maximum": None, "minimum": None}})
        assert len(extracted_features) == 2

    def test_extract_features_for_one_time_series(self):
        # todo: implement more methods and test more aspects
        df = self.create_test_data_sample()
        settings = ComprehensiveFCParameters()

        extracted_features = extract_features(df, default_fc_parameters=settings,
                                              column_value="val", column_id="id",
                                              column_kind="kind", column_sort="sort")

        self.assertIsInstance(extracted_features, pd.DataFrame)
        self.assertTrue(np.all(extracted_features.b__sum_values == np.array([757, 695])))
        self.assertTrue(np.all(extracted_features.b__minimum == np.array([3, 1])))
        self.assertTrue(np.all(extracted_features.b__abs_energy == np.array([36619, 35483])))
        self.assertTrue(np.all(extracted_features.b__mean == np.array([37.85, 34.75])))
        self.assertTrue(np.all(extracted_features.b__median == np.array([39.5, 28.0])))

        df_sts = self.create_one_valued_time_series()
        extracted_features_sts = extract_features(df_sts, default_fc_parameters=settings,
                                                  column_value="val", column_id="id",
                                                  column_kind="kind", column_sort="sort")

        self.assertIsInstance(extracted_features_sts, pd.DataFrame)
        self.assertTrue(np.all(extracted_features_sts.a__maximum == np.array([1.0, 6.0])))
        self.assertTrue(np.all(extracted_features_sts.a__sum_values == np.array([1.0, 11.0])))
        self.assertTrue(np.all(extracted_features_sts.a__count_above_mean == np.array([0, 1])))

    def test_extract_features_for_index_based_functions(self):
        df = self.create_test_data_sample_with_time_index()

        settings = {
            'linear_trend_timewise': [{"attr": "slope"}],
            'linear_trend': [{"attr": "slope"}]
        }

        extracted_features = extract_features(df, default_fc_parameters=settings,
                                              column_value="val", column_id="id",
                                              column_kind="kind",
                                              column_sort="sort")

        self.assertIsInstance(extracted_features, pd.DataFrame)

        slope_a = extracted_features['a__linear_trend_timewise__attr_"slope"'].values
        slope_b = extracted_features['b__linear_trend_timewise__attr_"slope"'].values

        self.assertAlmostEqual(slope_a[0], -0.001347117)
        self.assertAlmostEqual(slope_a[1], 0.052036340)
        self.assertAlmostEqual(slope_b[0], 0.021898496)
        self.assertAlmostEqual(slope_b[1], -0.012312)

        # Test that the index of the returned df is the ID and not the timestamp
        self.assertTrue(extracted_features.index.dtype != df.index.dtype)
        self.assertTrue(extracted_features.index.dtype == df['id'].dtype)
        self.assertEqual(
            sorted(extracted_features.index.unique().tolist()), sorted(df['id'].unique().tolist())
        )

    def test_extract_features_after_randomisation(self):
        df = self.create_test_data_sample()
        df_random = df.copy().sample(frac=1)

        extracted_features = extract_features(df, column_id="id", column_sort="sort",
                                              column_kind="kind",
                                              column_value="val",
                                              n_jobs=self.n_jobs).sort_index()
        extracted_features_from_random = extract_features(df_random, column_id="id",
                                                          column_sort="sort",
                                                          column_kind="kind",
                                                          column_value="val",
                                                          n_jobs=self.n_jobs).sort_index()

        self.assertCountEqual(extracted_features.columns,
                              extracted_features_from_random.columns)

        for col in extracted_features:
            self.assertIsNone(np.testing.assert_array_almost_equal(extracted_features[col],
                                                                   extracted_features_from_random[
                                                                       col]))

    def test_profiling_file_written_out(self):

        df = pd.DataFrame(data={"id": np.repeat([1, 2], 10), "val": np.random.normal(0, 1, 20)})
        profiling_filename = os.path.join(self.directory, "test_profiling.txt")
        X = extract_features(df, column_id="id", column_value="val", n_jobs=self.n_jobs,
                             profile=True, profiling_filename=profiling_filename)

        self.assertTrue(os.path.isfile(profiling_filename))
        os.remove(profiling_filename)

    def test_profiling_cumulative_file_written_out(self):

        PROFILING_FILENAME = os.path.join(self.directory, "test_profiling_cumulative.txt")
        PROFILING_SORTING = "cumulative"

        df = pd.DataFrame(data={"id": np.repeat([1, 2], 10), "val": np.random.normal(0, 1, 20)})
        extract_features(df, column_id="id", column_value="val", n_jobs=self.n_jobs,
                         profile=True, profiling_filename=PROFILING_FILENAME,
                         profiling_sorting=PROFILING_SORTING)

        self.assertTrue(os.path.isfile(PROFILING_FILENAME))
        os.remove(PROFILING_FILENAME)

    def test_extract_features_without_settings(self):
        df = pd.DataFrame(data={"id": np.repeat([1, 2], 10),
                                "value1": np.random.normal(0, 1, 20),
                                "value2": np.random.normal(0, 1, 20)})
        X = extract_features(df, column_id="id",
                             n_jobs=self.n_jobs)
        self.assertIn("value1__maximum", list(X.columns))
        self.assertIn("value2__maximum", list(X.columns))

    def test_extract_features_with_and_without_parallelization(self):
        df = self.create_test_data_sample()

        features_parallel = extract_features(df, column_id="id", column_sort="sort",
                                             column_kind="kind", column_value="val",
                                             n_jobs=self.n_jobs)

        features_serial = extract_features(df, column_id="id", column_sort="sort",
                                           column_kind="kind", column_value="val",
                                           n_jobs=0)

        self.assertCountEqual(features_parallel.columns, features_serial.columns)

        for col in features_parallel.columns:
            np.testing.assert_array_almost_equal(features_parallel[col], features_serial[col])

    def test_extract_index_preservation(self):
        df = self.create_test_data_nearly_numerical_indices()
        extracted_features = extract_features(df, column_id="id", column_sort="sort",
                                              column_kind="kind", column_value="val",
                                              n_jobs=self.n_jobs)

        self.assertIsInstance(extracted_features, pd.DataFrame)
        self.assertEqual(set(df["id"]), set(extracted_features.index))

    def test_extract_features_alphabetically_sorted(self):
        df = self.create_test_data_sample()

        features = extract_features(df, column_id="id", column_sort="sort",
                                    column_kind="kind", column_value="val")

        for col_name in features.columns:
            # split out the configuration of the features calculator
            col_name_chunks = col_name.split("__")
            # the name is always at the beginning, so remove it. Also remove the kind of the column
            col_name_chunks = col_name_chunks[2:]

            self.assertEqual(col_name_chunks, list(sorted(col_name_chunks)))


class ParallelExtractionTestCase(DataTestCase):
    def setUp(self):
        self.n_jobs = 2

        # only calculate some features to reduce load on travis ci
        self.name_to_param = {"maximum": None,
                              "sum_values": None,
                              "abs_energy": None,
                              "minimum": None,
                              "mean": None,
                              "median": None}

    def test_extract_features(self):
        # todo: implement more methods and test more aspects
        df = self.create_test_data_sample()
        extracted_features = extract_features(df, column_id="id", column_sort="sort",
                                              column_kind="kind",
                                              column_value="val",
                                              n_jobs=self.n_jobs)

        self.assertIsInstance(extracted_features, pd.DataFrame)
        self.assertTrue(np.all(extracted_features.a__maximum == np.array([71, 77])))
        self.assertTrue(np.all(extracted_features.a__sum_values == np.array([691, 1017])))
        self.assertTrue(np.all(extracted_features.a__abs_energy == np.array([32211, 63167])))
        self.assertTrue(np.all(extracted_features.b__sum_values == np.array([757, 695])))
        self.assertTrue(np.all(extracted_features.b__minimum == np.array([3, 1])))
        self.assertTrue(np.all(extracted_features.b__abs_energy == np.array([36619, 35483])))
        self.assertTrue(np.all(extracted_features.b__mean == np.array([37.85, 34.75])))
        self.assertTrue(np.all(extracted_features.b__median == np.array([39.5, 28.0])))


class DistributorUsageTestCase(DataTestCase):
    def setUp(self):
        # only calculate some features to reduce load on travis ci
        self.name_to_param = {"maximum": None}

    def test_distributor_map_reduce_is_called(self):
        df = self.create_test_data_sample()

        mock = Mock(spec=IterableDistributorBaseClass)
        mock.close.return_value = None
        mock.map_reduce.return_value = []

        X = extract_features(timeseries_container=df, column_id="id", column_sort="sort",
                             column_kind="kind", column_value="val",
                             default_fc_parameters=self.name_to_param, distributor=mock)

        self.assertTrue(mock.map_reduce.called)

    def test_distributor_close_is_called(self):
        df = self.create_test_data_sample()

        mock = MapDistributor()
        mock.close = Mock()
        mock.close.return_value = None

        X = extract_features(timeseries_container=df, column_id="id", column_sort="sort",
                             column_kind="kind", column_value="val",
                             default_fc_parameters=self.name_to_param, distributor=mock)

        self.assertTrue(mock.close.called)



import math

import numpy as np
import pandas as pd
from dask import dataframe as dd
from unittest import TestCase
from unittest.mock import Mock

from tests.fixtures import DataTestCase
from tsfresh.feature_extraction.data import (
    to_tsdata,
    LongTsFrameAdapter,
    WideTsFrameAdapter,
    TsDictAdapter,
    PartitionedTsData
)
from tsfresh.utilities.distribution import MultiprocessingDistributor

TEST_DATA_EXPECTED_TUPLES = \
    [(10, 'a', pd.Series([36, 71, 27, 62, 56, 58, 67, 11, 2, 24, 45, 30, 0,
                          9, 41, 28, 33, 19, 29, 43],
                         index=[10] * 20)),
     (10, 'b', pd.Series([78, 37, 23, 44, 6, 3, 21, 61, 39, 31, 53, 16, 66,
                          50, 40, 47, 7, 42, 38, 55],
                         index=[10] * 20)),
     (500, 'a', pd.Series([76, 72, 74, 75, 32, 64, 46, 35, 15, 70, 57, 65,
                           51, 26, 5, 25, 10, 69, 73, 77],
                          index=[500] * 20)),
     (500, 'b', pd.Series([8, 60, 12, 68, 22, 17, 18, 63, 49, 34, 20, 52,
                           48, 14, 79, 4, 1, 59, 54, 13],
                          index=[500] * 20))]

WIDE_TEST_DATA_EXPECTED_TUPLES = \
    [(10, 'a', pd.Series([11, 9, 67, 45, 30, 58, 62, 19, 56, 29, 0, 27, 36,
                          43, 33, 2, 24, 71, 41, 28],
                         index=list(range(20)))),
     (10, 'b', pd.Series([50, 40, 39, 7, 53, 23, 16, 37, 66, 38, 6, 47, 3,
                          61, 44, 42, 78, 31, 21, 55],
                         index=list(range(20)))),
     (500, 'a', pd.Series([15, 35, 25, 32, 69, 65, 70, 64, 51, 46, 5, 77,
                           26, 73, 76, 75, 72, 74, 10, 57],
                          index=list(range(20, 40)))),
     (500, 'b', pd.Series([4, 14, 68, 22, 18, 52, 54, 60, 79, 12, 49, 63,
                           8, 59, 1, 13, 20, 17, 48, 34],
                          index=list(range(20, 40))))]


class DataAdapterTestCase(DataTestCase):
    def test_long_tsframe(self):
        df = self.create_test_data_sample()
        data = LongTsFrameAdapter(df, "id", "kind", "val", "sort")

        self.assert_tsdata(data, TEST_DATA_EXPECTED_TUPLES)

    def test_long_tsframe_no_value_column(self):
        df = self.create_test_data_sample()
        data = LongTsFrameAdapter(df, "id", "kind", None, "sort")

        self.assert_tsdata(data, TEST_DATA_EXPECTED_TUPLES)

    def test_wide_tsframe(self):
        df = self.create_test_data_sample_wide()
        data = WideTsFrameAdapter(df, "id", "sort")

        self.assert_tsdata(data, WIDE_TEST_DATA_EXPECTED_TUPLES)

    def test_wide_tsframe_without_sort(self):
        df = self.create_test_data_sample_wide()
        del df["sort"]
        data = WideTsFrameAdapter(df, "id")

        self.assert_tsdata(data, WIDE_TEST_DATA_EXPECTED_TUPLES)

    def test_dict_tsframe(self):
        df = {key: df for key, df in self.create_test_data_sample().groupby(["kind"])}
        data = TsDictAdapter(df, "id", "val", "sort")

        self.assert_tsdata(data, TEST_DATA_EXPECTED_TUPLES)

    def assert_tsdata(self, data, expected):
        self.assertEqual(len(data), len(expected))
        self.assertEqual(sum(1 for _ in data), len(data))
        self.assert_data_chunk_object_equal(data, expected)

    def assert_data_chunk_object_equal(self, result, expected):
        dic_result = {str(x[0]) + "_" + str(x[1]): x[2] for x in result}
        dic_expected = {str(x[0]) + "_" + str(x[1]): x[2] for x in expected}
        for k in dic_result.keys():
            pd.testing.assert_series_equal(dic_result[k], dic_expected[k], check_names=False)

    def test_simple_data_sample_two_timeseries(self):
        df = pd.DataFrame({"id": [10] * 4, "kind": ["a"] * 2 + ["b"] * 2, "val": [36, 71, 78, 37]})
        df.set_index("id", drop=False, inplace=True)
        df.index.name = None

        result = to_tsdata(df, "id", "kind", "val")
        expected = [(10, 'a', pd.Series([36, 71], index=[10] * 2, name="val")),
                    (10, 'b', pd.Series([78, 37], index=[10] * 2, name="val"))]
        self.assert_data_chunk_object_equal(result, expected)

    def test_simple_data_sample_four_timeseries(self):
        df = self.create_test_data_sample()
        df.index.name = None
        df.sort_values(by=["id", "kind", "sort"], inplace=True)

        result = to_tsdata(df, "id", "kind", "val", "sort")
        expected = TEST_DATA_EXPECTED_TUPLES

        self.assert_data_chunk_object_equal(result, expected)

    def test_with_dictionaries_two_rows(self):
        test_df = pd.DataFrame([{"value": 2, "sort": 2, "id": "id_1"},
                                {"value": 1, "sort": 1, "id": "id_1"}])
        test_dict = {"a": test_df, "b": test_df}

        result = to_tsdata(test_dict, column_id="id", column_value="value", column_sort="sort")
        expected = [("id_1", 'a', pd.Series([1, 2], index=[1, 0], name="value")),
                    ("id_1", 'b', pd.Series([1, 2], index=[1, 0], name="value"))]
        self.assert_data_chunk_object_equal(result, expected)

    def test_with_dictionaries_two_rows(self):
        test_df = pd.DataFrame([{"value": 1, "id": "id_1"},
                                {"value": 2, "id": "id_1"}])
        test_dict = {"a": test_df, "b": test_df}

        result = to_tsdata(test_dict, column_id="id", column_value="value")
        expected = [("id_1", 'a', pd.Series([1, 2], index=[0, 1], name="value")),
                    ("id_1", 'b', pd.Series([1, 2], index=[0, 1], name="value"))]
        self.assert_data_chunk_object_equal(result, expected)

    def test_wide_dataframe_order_preserved_with_sort_column(self):
        """ verifies that the order of the sort column from a wide time series container is preserved
        """

        test_df = pd.DataFrame({'id': ["a", "a", "b"],
                                'v1': [3, 2, 1],
                                'v2': [13, 12, 11],
                                'sort': [103, 102, 101]})

        result = to_tsdata(test_df, column_id="id", column_sort="sort")
        expected = [("a", 'v1', pd.Series([2, 3], index=[1, 0], name="v1")),
                    ("a", 'v2', pd.Series([12, 13], index=[1, 0], name="v2")),
                    ("b", 'v1', pd.Series([1], index=[2], name="v1")),
                    ("b", 'v2', pd.Series([11], index=[2], name="v2"))]
        self.assert_data_chunk_object_equal(result, expected)

    def test_dask_dataframe_with_kind(self):
        test_df = dd.from_pandas(pd.DataFrame({
            "id": [1, 2],
            "kind": ["a", "a"],
            "value": [1, 2]
        }), npartitions=1)

        result = to_tsdata(test_df, column_id="id", column_kind="kind")
        self.assertEqual(result.column_id, "id")
        self.assertEqual(result.column_kind, "kind")
        self.assertEqual(result.column_value, "value")

        def test_f(chunk):
            return pd.DataFrame({"id": chunk[0], "variable": chunk[1], "value": chunk[2]})

        return_f = result.apply(test_f, meta=(("id", "int"), ("variable", "int"), ("value", "int"))).compute()
        pd.testing.assert_frame_equal(return_f, pd.DataFrame({
            "id": [1, 2],
            "variable": ["a", "a"],
            "value": [1.0, 2.0]
        }))

    def test_dask_dataframe_without_kind(self):
        test_df = dd.from_pandas(pd.DataFrame({
            "id": [1, 2],
            "value_a": [1, 2],
            "value_b": [3, 4]
        }), npartitions=1)

        result = to_tsdata(test_df, column_id="id")
        self.assertEqual(result.column_id, "id")

        def test_f(chunk):
            return pd.DataFrame({"id": chunk[0], "variable": chunk[1], "value": chunk[2]})

        return_f = result.apply(test_f, meta=(("id", "int"), ("variable", "int"), ("value", "int"))).compute()
        pd.testing.assert_frame_equal(return_f.reset_index(drop=True), pd.DataFrame({
            "id": [1, 2, 1, 2],
            "variable": ["value_a", "value_a", "value_b", "value_b"],
            "value": [1.0, 2.0, 3.0, 4.0]
        }))

        test_df = dd.from_pandas(pd.DataFrame({
            "id": [1, 1],
            "sort": [2, 1],
            "value_a": [1, 2],
            "value_b": [3, 4]
        }), npartitions=1)

        result = to_tsdata(test_df, column_id="id", column_sort="sort")
        self.assertEqual(result.column_id, "id")

        def test_f(chunk):
            return pd.DataFrame({"id": chunk[0], "variable": chunk[1], "value": chunk[2]})

        return_f = result.apply(test_f, meta=(("id", "int"), ("variable", "int"), ("value", "int"))).compute()

        pd.testing.assert_frame_equal(return_f.reset_index(drop=True), pd.DataFrame({
            "id": [1, 1, 1, 1],
            "variable": ["value_a", "value_a", "value_b", "value_b"],
            "value": [2.0, 1.0, 4.0, 3.0]
        }))

    def test_with_wrong_input(self):
        test_df = pd.DataFrame([{"id": 0, "kind": "a", "value": 3, "sort": np.NaN}])
        self.assertRaises(ValueError, to_tsdata, test_df,
                          "id", "kind", "value", "sort")

        test_df = pd.DataFrame([{"id": 0, "kind": "a", "value": 3, "sort": 1}])
        self.assertRaises(ValueError, to_tsdata, test_df,
                          "strange_id", "kind", "value", "sort")
        test_df = dd.from_pandas(test_df, npartitions=1)
        self.assertRaises(ValueError, to_tsdata, test_df,
                          "strange_id", "kind", "value", "sort")

        test_df = pd.DataFrame([{"id": 0, "kind": "a", "value": 3, "value_2": 1, "sort": 1}])
        self.assertRaises(ValueError, to_tsdata, test_df,
                          "strange_id", "kind", None, "sort")
        test_df = dd.from_pandas(test_df, npartitions=1)
        self.assertRaises(ValueError, to_tsdata, test_df,
                          "strange_id", "kind", None, "sort")

        test_df = pd.DataFrame([{"id": 0, "kind": "a", "value": 3, "sort": 1}])
        self.assertRaises(ValueError, to_tsdata, test_df,
                          "id", "strange_kind", "value", "sort")
        test_df = dd.from_pandas(test_df, npartitions=1)
        self.assertRaises(ValueError, to_tsdata, test_df,
                          "id", "strange_kind", "value", "sort")

        test_df = pd.DataFrame([{"id": np.NaN, "kind": "a", "value": 3, "sort": 1}])
        self.assertRaises(ValueError, to_tsdata, test_df,
                          "id", "kind", "value", "sort")

        test_df = pd.DataFrame([{"id": 0, "kind": np.NaN, "value": 3, "sort": 1}])
        self.assertRaises(ValueError, to_tsdata, test_df,
                          "id", "kind", "value", "sort")

        test_df = pd.DataFrame([{"id": 2}, {"id": 1}])
        test_dd = dd.from_pandas(test_df, npartitions=1)
        test_dict = {"a": test_df, "b": test_df}

        # column_id needs to be given
        self.assertRaises(ValueError, to_tsdata, test_df,
                          None, "a", "b", None)
        self.assertRaises(ValueError, to_tsdata, test_dd,
                          None, "a", "b", None)
        self.assertRaises(ValueError, to_tsdata, test_df,
                          None, "a", "b", "a")
        self.assertRaises(ValueError, to_tsdata, test_dd,
                          None, "a", "b", "a")
        self.assertRaises(ValueError, to_tsdata, test_dict,
                          None, "a", "b", None)
        self.assertRaises(ValueError, to_tsdata, test_dict,
                          None, "a", "b", "a")

        # If there are more than one column, the algorithm can not choose the correct column
        self.assertRaises(ValueError, to_tsdata, test_dict,
                          "id", None, None, None)

        test_dict = {"a": pd.DataFrame([{"id": 2, "value_a": 3}, {"id": 1, "value_a": 4}]),
                     "b": pd.DataFrame([{"id": 2}, {"id": 1}])}

        # If there are more than one column, the algorithm can not choose the correct column
        self.assertRaises(ValueError, to_tsdata, test_dict,
                          "id", None, None, None)

        test_df = pd.DataFrame([{"id": 0, "value": np.NaN}])
        self.assertRaises(ValueError, to_tsdata, test_df,
                          "id", None, "value", None)

        test_df = pd.DataFrame([{"id": 0, "value": np.NaN}])
        self.assertRaises(ValueError, to_tsdata, test_df,
                          None, None, "value", None)

        test_df = pd.DataFrame([{"id": 0, "a_": 3, "b": 5, "sort": 1}])
        self.assertRaises(ValueError, to_tsdata, test_df,
                          "id", None, None, "sort")
        test_df = dd.from_pandas(test_df, npartitions=1)
        self.assertRaises(ValueError, to_tsdata, test_df,
                          "id", None, None, "sort")

        test_df = pd.DataFrame([{"id": 0, "a__c": 3, "b": 5, "sort": 1}])
        self.assertRaises(ValueError, to_tsdata, test_df,
                          "id", None, None, "sort")
        test_df = dd.from_pandas(test_df, npartitions=1)
        self.assertRaises(ValueError, to_tsdata, test_df,
                          "id", None, None, "sort")

        test_df = pd.DataFrame([{"id": 0}])
        self.assertRaises(ValueError, to_tsdata, test_df,
                          "id", None, None, None)
        test_df = dd.from_pandas(test_df, npartitions=1)
        self.assertRaises(ValueError, to_tsdata, test_df,
                          "id", None, None, None)

        test_df = pd.DataFrame([{"id": 0, "sort": 0}])
        self.assertRaises(ValueError, to_tsdata, test_df,
                          "id", None, None, "sort")
        test_df = dd.from_pandas(test_df, npartitions=1)
        self.assertRaises(ValueError, to_tsdata, test_df,
                          "id", None, None, "sort")

        test_df = [1, 2, 3]
        self.assertRaises(ValueError, to_tsdata, test_df,
                          "a", "b", "c", "d")


class PivotListTestCase(TestCase):
    def test_empty_list(self):
        mock_ts_data = Mock()
        mock_ts_data.df_id_type = str

        return_df = PartitionedTsData.pivot(mock_ts_data, [])

        self.assertEqual(len(return_df), 0)
        self.assertEqual(len(return_df.index), 0)
        self.assertEqual(len(return_df.columns), 0)

    def test_different_input(self):
        mock_ts_data = Mock()
        mock_ts_data.df_id_type = str

        input_list = [
            ("a", "b", 1),
            ("a", "c", 2),
            ("A", "b", 3),
            ("A", "c", 4),
            ("X", "Y", 5),
        ]
        return_df = PartitionedTsData.pivot(mock_ts_data, input_list)

        self.assertEqual(len(return_df), 3)
        self.assertEqual(set(return_df.index), {"a", "A", "X"})
        self.assertEqual(set(return_df.columns), {"b", "c", "Y"})

        self.assertEqual(return_df.loc["a", "b"], 1)
        self.assertEqual(return_df.loc["a", "c"], 2)
        self.assertEqual(return_df.loc["A", "b"], 3)
        self.assertEqual(return_df.loc["A", "c"], 4)
        self.assertEqual(return_df.loc["X", "Y"], 5)

    def test_long_input(self):
        mock_ts_data = Mock()
        mock_ts_data.df_id_type = str

        input_list = []
        for i in range(100):
            for j in range(100):
                input_list.append((i, j, i*j))

        return_df = PartitionedTsData.pivot(mock_ts_data, input_list)

        self.assertEqual(len(return_df), 100)
        self.assertEqual(len(return_df.columns), 100)
        # every cell should be filled
        self.assertEqual(np.sum(np.sum(np.isnan(return_df))), 0)
        # test if all entries are there
        self.assertEqual(return_df.sum().sum(), 24502500)



# -*- coding: utf-8 -*-
# This file as well as the whole tsfresh package are licenced under the MIT licence (see the LICENCE.txt)
# Maximilian Christ (maximilianchrist.com), Blue Yonder Gmbh, 2017
from glob import glob
import time
import pandas as pd
from subprocess import check_output


def test_with_length(length, df):
    from tsfresh import extract_features

    start = time.time()
    df = extract_features(df[:length], column_id='id', column_sort='time')
    end = time.time()

    duration = end - start

    print("Some checks with length", length)
    print(100 * duration)
    print(len(df.columns), len(df))
    print(df[["a__abs_energy", "b__absolute_sum_of_changes", "f__time_reversal_asymmetry_statistic__lag_1"]].head())

    return {"length": length, "duration": duration}


def plot_results():
    from matplotlib import pyplot as plt

    plt.figure(figsize=(7, 7))

    baseline = pd.read_csv("a57a09fe62a62fe0d2564a056f7fd99f58822312.dat").groupby("length").duration.mean()

    for file_name in glob("*.dat"):
        df = pd.read_csv(file_name).groupby("length").duration.mean()

        plt.subplot(211)
        df.plot(label=file_name.replace(".dat", ""))

        plt.subplot(212)
        (baseline / df).plot(label=file_name.replace(".dat", ""))

    plt.subplot(211)
    plt.xlabel("DataFrame Length")
    plt.ylabel("Extract Features Mean Duration")
    plt.legend()

    plt.subplot(212)
    plt.xlabel("DataFrame Length")
    plt.ylabel("Speedup")
    plt.gca().axhline(1, color="black", ls="--")
    plt.legend()

    plt.savefig("timing.png")


def test_timing():
    from tsfresh.examples.robot_execution_failures import download_robot_execution_failures, \
        load_robot_execution_failures
    download_robot_execution_failures()
    df, y = load_robot_execution_failures()

    commit_hash = check_output(["git", "log", "--format=\"%H\"", "-1"]).decode("ascii").strip().replace("\"", "")

    lengths_to_test = [1, 5, 10, 60, 100, 400, 600, 1000, 2000]
    results = []

    for length in lengths_to_test:
        results.append(test_with_length(length, df))
        results.append(test_with_length(length, df))
        results.append(test_with_length(length, df))

    results = pd.DataFrame(results)
    results.to_csv("{hash}.dat".format(hash=commit_hash))


if __name__ == "__main__":
    test_timing()
    plot_results()



from unittest import TestCase

from tsfresh.convenience.bindings import dask_feature_extraction_on_chunk
from tsfresh.feature_extraction.settings import MinimalFCParameters

from dask import dataframe as dd
import pandas as pd


class DaskBindingsTestCase(TestCase):
    def test_feature_extraction(self):
        df = pd.DataFrame({"my_id": [1, 1, 1, 2, 2, 2], "my_kind": ["a"]*6,
                           "my_value": [1, 2, 3, 4, 5, 6]})

        df = dd.from_pandas(df, chunksize=3)

        df_grouped = df.groupby(["my_id", "my_kind"])

        features = dask_feature_extraction_on_chunk(df_grouped, column_id="my_id",
                                                    column_kind="my_kind",
                                                    column_value="my_value",
                                                    column_sort=None,
                                                    default_fc_parameters=MinimalFCParameters())

        features = features.compute()

        self.assertEqual(list(sorted(features.columns)), ["my_id", "value", "variable"])
        self.assertEqual(len(features), 2 * len(MinimalFCParameters()))



from functools import partial

from tsfresh.feature_extraction.extraction import _do_extraction_on_chunk
from tsfresh.feature_extraction.settings import ComprehensiveFCParameters

import pandas as pd


def _feature_extraction_on_chunk_helper(df, column_id, column_kind,
                                        column_sort, column_value,
                                        default_fc_parameters, kind_to_fc_parameters):
    """
    Helper function wrapped around _do_extraction_on_chunk to use the correct format
    of the "chunk" and output a pandas dataframe.
    Is used e.g. in the convenience functions for dask and spark.

    For the definitions of the parameters, please see these convenience functions.
    """
    if default_fc_parameters is None and kind_to_fc_parameters is None:
        default_fc_parameters = ComprehensiveFCParameters()
    elif default_fc_parameters is None and kind_to_fc_parameters is not None:
        default_fc_parameters = {}

    if column_sort is not None:
        df = df.sort_values(column_sort)

    chunk = df[column_id].iloc[0], df[column_kind].iloc[0], df[column_value]
    features = _do_extraction_on_chunk(chunk, default_fc_parameters=default_fc_parameters,
                                       kind_to_fc_parameters=kind_to_fc_parameters)
    features = pd.DataFrame(features, columns=[column_id, "variable", "value"])
    features["value"] = features["value"].astype("double")

    return features[[column_id, "variable", "value"]]


def dask_feature_extraction_on_chunk(df, column_id, column_kind,
                                     column_value, column_sort=None,
                                     default_fc_parameters=None, kind_to_fc_parameters=None):
    """
    Extract features on a grouped dask dataframe given the column names and the extraction settings.
    This wrapper function should only be used if you have a dask dataframe as input.
    All format handling (input and output) needs to be done before or after that.

    Examples
    ========

    For example if you want to extract features on the robot example dataframe (stored as csv):

    Import statements:

    >>>  from dask import dataframe as dd
    >>>  from tsfresh.convenience.bindings import dask_feature_extraction_on_chunk
    >>>  from tsfresh.feature_extraction.settings import MinimalFCParameters

    Read in the data

    >>>  df = dd.read_csv("robot.csv")

    Prepare the data into correct format.
    The format needs to be a grouped dataframe (grouped by time series id and feature kind),
    where each group chunk consists of a dataframe with exactly 4 columns: ``column_id``,
    ``column_kind``, ``column_sort`` and ``column_value``.
    You can find the description of the columns in :ref:`data-formats-label`.
    Please note: for this function to work you need to have all columns present!
    If necessary create the columns and fill them with dummy values.

    >>>  df = df.melt(id_vars=["id", "time"],
    ...               value_vars=["F_x", "F_y", "F_z", "T_x", "T_y", "T_z"],
    ...               var_name="kind", value_name="value")
    >>>  df_grouped = df.groupby(["id", "kind"])

    Call the feature extraction

    >>>  features = dask_feature_extraction_on_chunk(df_grouped, column_id="id", column_kind="kind",
    ...                                              column_sort="time", column_value="value",
    ...                                              default_fc_parameters=MinimalFCParameters())

    Write out the data in a tabular format

    >>>  features = features.categorize(columns=["variable"])
    >>>  features = features.reset_index(drop=True) \\
    ...                 .pivot_table(index="id", columns="variable", values="value", aggfunc="mean")
    >>>  features.to_csv("output")


    :param df: A dask dataframe grouped by id and kind.
    :type df: dask.dataframe.groupby.DataFrameGroupBy

    :param default_fc_parameters: mapping from feature calculator names to parameters. Only those names
           which are keys in this dict will be calculated. See the class:`ComprehensiveFCParameters` for
           more information.
    :type default_fc_parameters: dict

    :param kind_to_fc_parameters: mapping from kind names to objects of the same type as the ones for
            default_fc_parameters. If you put a kind as a key here, the fc_parameters
            object (which is the value), will be used instead of the default_fc_parameters. This means
            that kinds, for which kind_of_fc_parameters doe not have any entries, will be ignored by
            the feature selection.
    :type kind_to_fc_parameters: dict

    :param column_id: The name of the id column to group by.
    :type column_id: str

    :param column_sort: The name of the sort column.
    :type column_sort: str or None

    :param column_kind: The name of the column keeping record on the kind of the value.
    :type column_kind: str

    :param column_value: The name for the column keeping the value itself.
    :type column_value: str

    :return: A dask dataframe with the columns ``column_id``, "variable" and "value". The index is taken
            from the grouped dataframe.
    :rtype: dask.dataframe.DataFrame (id int64, variable object, value float64)

    """
    feature_extraction = partial(_feature_extraction_on_chunk_helper,
                                 column_id=column_id, column_kind=column_kind,
                                 column_sort=column_sort, column_value=column_value,
                                 default_fc_parameters=default_fc_parameters,
                                 kind_to_fc_parameters=kind_to_fc_parameters)
    return df.apply(feature_extraction, meta=[(column_id, 'int64'), ('variable', 'object'), ('value', 'float64')])


def spark_feature_extraction_on_chunk(df, column_id, column_kind,
                                      column_value, column_sort=None,
                                      default_fc_parameters=None, kind_to_fc_parameters=None):
    """
    Extract features on a grouped spark dataframe given the column names and the extraction settings.
    This wrapper function should only be used if you have a spark dataframe as input.
    All format handling (input and output) needs to be done before or after that.

    Examples
    ========

    For example if you want to extract features on the robot example dataframe (stored as csv):

    Import statements:

    >>>  from tsfresh.convenience.bindings import spark_feature_extraction_on_chunk
    >>>  from tsfresh.feature_extraction.settings import MinimalFCParameters

    Read in the data

    >>>  df = spark.read(...)

    Prepare the data into correct format.
    The format needs to be a grouped dataframe (grouped by time series id and feature kind),
    where each group chunk consists of a dataframe with exactly 4 columns: ``column_id``,
    ``column_kind``, ``column_sort`` and ``column_value``.
    You can find the description of the columns in :ref:`data-formats-label`.
    Please note: for this function to work you need to have all columns present!
    If necessary create the columns and fill them with dummy values.

    >>>  df = ...
    >>>  df_grouped = df.groupby(["id", "kind"])

    Call the feature extraction

    >>>  features = spark_feature_extraction_on_chunk(df_grouped, column_id="id", column_kind="kind",
    ...                                               column_sort="time", column_value="value",
    ...                                               default_fc_parameters=MinimalFCParameters())

    Write out the data in a tabular format

    >>>  features = features.groupby("id").pivot("variable").sum("value")
    >>>  features.write.csv("output")


    :param df: A spark dataframe grouped by id and kind.
    :type df: pyspark.sql.group.GroupedData

    :param default_fc_parameters: mapping from feature calculator names to parameters. Only those names
           which are keys in this dict will be calculated. See the class:`ComprehensiveFCParameters` for
           more information.
    :type default_fc_parameters: dict

    :param kind_to_fc_parameters: mapping from kind names to objects of the same type as the ones for
            default_fc_parameters. If you put a kind as a key here, the fc_parameters
            object (which is the value), will be used instead of the default_fc_parameters.
            This means that kinds, for which kind_of_fc_parameters doe not have any entries,
            will be ignored by the feature selection.
    :type kind_to_fc_parameters: dict

    :param column_id: The name of the id column to group by.
    :type column_id: str

    :param column_sort: The name of the sort column.
    :type column_sort: str or None

    :param column_kind: The name of the column keeping record on the kind of the value.
    :type column_kind: str

    :param column_value: The name for the column keeping the value itself.
    :type column_value: str

    :return: A dask dataframe with the columns ``column_id``, "variable" and "value".
    :rtype: pyspark.sql.DataFrame[id: bigint, variable: string, value: double]

    """
    from pyspark.sql.functions import pandas_udf, PandasUDFType

    feature_extraction = partial(_feature_extraction_on_chunk_helper,
                                 column_id=column_id, column_kind=column_kind,
                                 column_sort=column_sort, column_value=column_value,
                                 default_fc_parameters=default_fc_parameters,
                                 kind_to_fc_parameters=kind_to_fc_parameters)

    type_string = "{column_id} long, variable string, value double".format(column_id=column_id)
    feature_extraction_udf = pandas_udf(type_string,
                                        PandasUDFType.GROUPED_MAP)(feature_extraction)

    return df.apply(feature_extraction_udf)
